
import tkinter as tk
import tkinter.ttk
import tkinter.simpledialog
import tkinter.messagebox
from sys import maxsize
from collections import OrderedDict
from .dialog_elements import (
    FileEntry,
    IntegerEntry,
    Checkbox,
    StringEntry,
    SectionLabel,
    DEFAULT_COLUMNS,
)
from ..basic import BasicSeqLib
from ..barcodevariant import BcvSeqLib
from ..barcodeid import BcidSeqLib
from ..barcode import BarcodeSeqLib
from ..idonly import IdOnlySeqLib
from ..overlap import OverlapSeqLib
from ..seqlib import SeqLib
from ..variant import VariantSeqLib


def clear_nones_filter(v):
    """
    Filter function for clear_nones.

    Returns False if v is None, else True.
    """
    if isinstance(v, dict):
        if len(list(v.keys())) == 0:
            # removing empty dictionaries breaks SeqLib recognition
            # return False
            return True
        else:
            return True
    elif v is None:
        return False
    else:
        return True


def clear_nones(d):
    """
    Return a copy of dictionary *d* with keys with value ``None`` removed.
    """
    if not isinstance(d, dict):
        return d
    else:
        return dict(
            (k, clear_nones(v)) for k, v in d.items() if clear_nones_filter(v)
        )


# All valid suffixes for a FASTQ file that can be recognized by Enrich2
_FASTQ_SUFFIXES = [x + y for x in (".fq", ".fastq") for y in ("", ".bz2", ".gz")]


#: Dictionary defining the layout of the edit UI elements in columns
#: Keys are class names and values are lists of tuples, one tuple per column
#: The column tuples contain keys to the dialog element's ``frame_dict`` member
element_layouts = {
    "BcvSeqLib": [
        ("main", "counts"),
        ("fastq", "trimming", "filters"),
        ("barcodes", "variants"),
    ],
    "BcidSeqLib": [
        ("main", "counts"),
        ("fastq", "trimming", "filters"),
        ("barcodes", "identifiers"),
    ],
    "OverlapSeqLib": [
        ("main", "counts"),
        ("fastq", "filters", "overlap"),
        ("variants",),
    ],
    "BasicSeqLib": [
        ("main", "counts"),
        ("fastq", "trimming", "filters"),
        ("variants",),
    ],
    "BarcodeSeqLib": [
        ("main", "counts"),
        ("fastq", "trimming", "filters"),
        ("barcodes",),
    ],
    "IdOnlySeqLib": [("main", "counts"), ("identifiers",)],
    "Selection": [("main",)],
    "Condition": [("main",)],
    "Experiment": [("main",)],
}


class CountsToggle(object):
    def __init__(self, frame_dict):
        self.mode = tk.StringVar()
        self.frame_dict = frame_dict
        self.rb_fastq = None
        self.rb_coutns = None

    def body(self, master, row, columns=DEFAULT_COLUMNS, **kwargs):
        self.rb_fastq = tkinter.ttk.Radiobutton(
            master,
            text="FASTQ File Mode",
            variable=self.mode,
            value="FASTQ",
            command=self.fastq_mode,
        )
        self.rb_fastq.grid(row=row, column=0, columnspan=columns, sticky="ew")
        self.rb_counts = tkinter.ttk.Radiobutton(
            master,
            text="Count File Mode",
            variable=self.mode,
            value="Counts",
            command=self.counts_mode,
        )
        self.rb_counts.grid(row=row + 1, column=0, columnspan=columns, sticky="ew")
        return 2

    def counts_mode(self):
        for k in ["filters", "fastq", "overlap", "trimming"]:
            if k in self.frame_dict:
                for x in self.frame_dict[k]:
                    try:
                        x.disable()
                    except AttributeError:
                        print(x, k)
        for k in ["counts"]:
            if k in self.frame_dict:
                for x in self.frame_dict[k]:
                    try:
                        x.enable()
                    except AttributeError:
                        print(x, k)

    def fastq_mode(self):
        for k in ["counts"]:
            if k in self.frame_dict:
                for x in self.frame_dict[k]:
                    try:
                        x.disable()
                    except AttributeError:
                        print(x, k)
        for k in ["filters", "fastq", "overlap", "trimming"]:
            if k in self.frame_dict:
                for x in self.frame_dict[k]:
                    try:
                        x.enable()
                    except AttributeError:
                        print(x, k)

    def validate(self):
        return True

    def apply(self):
        return None

    def enable(self):
        pass

    def disable(self):
        pass


class EditDialog(tkinter.simpledialog.Dialog):
    """
    Dialog box for editing elements. Also used to set properties on newly-created elements.

    *parent_window* is the Tk window that owns this child window
    *tree* is the object containing the config tree and associated Treeview
    *new* is ``True`` if we are creating a new child of the focused item or ``False`` if we are editing the focused item
    """

    def __init__(self, parent_window, tree, element, title="Configure Object"):
        self.tree = tree
        self.element = element
        self.element_cfg = None
        self.frame_dict = OrderedDict()
        self.toggle = None

        # create the editable version of the config object
        self.element_cfg = self.element.serialize()

        # dialog options common to all elements
        self.frame_dict["main"] = list()
        self.name_entry = StringEntry("Name", self.element_cfg, "name", optional=False)
        self.frame_dict["main"].append(self.name_entry)
        if "output directory" in self.element_cfg:
            self.frame_dict["main"].append(
                FileEntry(
                    "Output Directory",
                    self.element_cfg,
                    "output directory",
                    optional=self.element != self.tree.root_element,
                    directory=True,
                )
            )
        if isinstance(self.element, SeqLib):
            self.frame_dict["counts"] = list()

            self.frame_dict["main"].append(SectionLabel("SeqLib Options"))
            self.frame_dict["main"].append(
                IntegerEntry("Time Point", self.element_cfg, "timepoint")
            )

            self.frame_dict["counts"].append(SectionLabel("Counts Options"))
            self.frame_dict["counts"].append(
                FileEntry(
                    "Counts File",
                    self.element_cfg,
                    "counts file",
                    extensions=[".h5", ".txt", ".tsv", ".csv"],
                )
            )

            if not isinstance(self.element, IdOnlySeqLib):
                self.toggle = CountsToggle(self.frame_dict)
                self.frame_dict["main"].append(self.toggle)

                self.frame_dict["fastq"] = list()
                self.frame_dict["filters"] = list()

                self.frame_dict["filters"].append(SectionLabel("FASTQ Filtering"))
                self.frame_dict["filters"].append(
                    IntegerEntry(
                        "Minimum Quality",
                        self.element_cfg["fastq"]["filters"],
                        "min quality",
                        optional=True,
                    )
                )
                self.frame_dict["filters"].append(
                    IntegerEntry(
                        "Average Quality",
                        self.element_cfg["fastq"]["filters"],
                        "avg quality",
                        optional=True,
                    )
                )
                self.frame_dict["filters"].append(
                    IntegerEntry(
                        "Maximum N's",
                        self.element_cfg["fastq"]["filters"],
                        "max N",
                        optional=True,
                    )
                )

                self.frame_dict["fastq"].append(SectionLabel("FASTQ Options"))

            if isinstance(self.element, OverlapSeqLib):
                self.frame_dict["fastq"].append(
                    FileEntry(
                        "Forward Reads",
                        self.element_cfg["fastq"],
                        "forward reads",
                        extensions=_FASTQ_SUFFIXES,
                    )
                )
                self.frame_dict["fastq"].append(
                    FileEntry(
                        "Reverse Reads",
                        self.element_cfg["fastq"],
                        "reverse reads",
                        extensions=_FASTQ_SUFFIXES,
                    )
                )
                self.frame_dict["overlap"] = list()
                self.frame_dict["overlap"].append(
                    IntegerEntry(
                        "Forward Start",
                        self.element_cfg["overlap"],
                        "forward start",
                        minvalue=1,
                    )
                )
                self.frame_dict["overlap"].append(
                    IntegerEntry(
                        "Reverse Start",
                        self.element_cfg["overlap"],
                        "reverse start",
                        minvalue=1,
                    )
                )
                self.frame_dict["overlap"].append(
                    IntegerEntry(
                        "Overlap Length",
                        self.element_cfg["overlap"],
                        "length",
                        minvalue=1,
                    )
                )
                self.frame_dict["overlap"].append(
                    IntegerEntry(
                        "Maximum Mismatches",
                        self.element_cfg["overlap"],
                        "max mismatches",
                    )
                )
                self.frame_dict["overlap"].append(
                    Checkbox("Overlap Only", self.element_cfg["overlap"], "trim")
                )
                self.frame_dict["filters"].append(
                    Checkbox(
                        "Remove Unresolvable Overlaps",
                        self.element_cfg["fastq"]["filters"],
                        "remove unresolvable",
                    )
                )
            elif "fastq" in self.frame_dict:
                self.frame_dict["fastq"].append(
                    FileEntry(
                        "Reads",
                        self.element_cfg["fastq"],
                        "reads",
                        extensions=_FASTQ_SUFFIXES,
                    )
                )
                self.frame_dict["fastq"].append(
                    Checkbox("Reverse", self.element_cfg["fastq"], "reverse")
                )

            if isinstance(self.element, BarcodeSeqLib) or isinstance(
                self.element, BasicSeqLib
            ):
                self.frame_dict["trimming"] = list()
                self.frame_dict["trimming"].append(
                    SectionLabel("Read Trimming Options")
                )
                self.frame_dict["trimming"].append(
                    IntegerEntry(
                        "Trim Start",
                        self.element_cfg["fastq"],
                        "start",
                        optional=True,
                        minvalue=1,
                    )
                )
                self.frame_dict["trimming"].append(
                    IntegerEntry(
                        "Trim Length",
                        self.element_cfg["fastq"],
                        "length",
                        optional=True,
                        minvalue=1,
                    )
                )

            if isinstance(self.element, BarcodeSeqLib):
                self.frame_dict["barcodes"] = list()
                self.frame_dict["barcodes"].append(SectionLabel("Barcode Options"))
                if isinstance(self.element, BcvSeqLib) or isinstance(
                    self.element, BcidSeqLib
                ):
                    self.frame_dict["barcodes"].append(
                        FileEntry(
                            "Barcode-variant File",
                            self.element_cfg["barcodes"],
                            "map file",
                        )
                    )
                self.frame_dict["barcodes"].append(
                    IntegerEntry(
                        "Minimum Count",
                        self.element_cfg["barcodes"],
                        "min count",
                        optional=True,
                    )
                )

            if isinstance(self.element, BcidSeqLib) or isinstance(
                self.element, IdOnlySeqLib
            ):
                self.frame_dict["identifiers"] = list()
                self.frame_dict["identifiers"].append(
                    SectionLabel("Identifier Options")
                )
                self.frame_dict["identifiers"].append(
                    IntegerEntry(
                        "Minimum Count",
                        self.element_cfg["identifiers"],
                        "min count",
                        optional=True,
                    )
                )

            if isinstance(self.element, VariantSeqLib):
                self.frame_dict["variants"] = list()
                self.frame_dict["variants"].append(SectionLabel("Variant Options"))
                self.frame_dict["variants"].append(
                    StringEntry(
                        "Wild Type Sequence",
                        self.element_cfg["variants"]["wild type"],
                        "sequence",
                    )
                )
                self.frame_dict["variants"].append(
                    IntegerEntry(
                        "Wild Type Offset",
                        self.element_cfg["variants"]["wild type"],
                        "reference offset",
                        optional=True,
                        minvalue=-maxsize,
                    )
                )
                self.frame_dict["variants"].append(
                    Checkbox(
                        "Protein Coding",
                        self.element_cfg["variants"]["wild type"],
                        "coding",
                    )
                )
                self.frame_dict["variants"].append(
                    IntegerEntry(
                        "Minimum Count",
                        self.element_cfg["variants"],
                        "min count",
                        optional=True,
                    )
                )
                self.frame_dict["variants"].append(
                    IntegerEntry(
                        "Maximum Mutations",
                        self.element_cfg["variants"],
                        "max mutations",
                        optional=True,
                    )
                )
                self.frame_dict["variants"].append(
                    Checkbox("Use Aligner", self.element_cfg["variants"], "use aligner")
                )

        tkinter.simpledialog.Dialog.__init__(self, parent_window, title)

    def body(self, master):
        """
        Add the UI elements to the edit window. Ordering and placement of UI 
        elements in columns is defined by the ``element_layouts`` dictionary.
        """
        main = tkinter.ttk.Frame(master, padding=(3, 3, 12, 12))
        main.grid(row=0, column=0, sticky="nsew")

        layout = element_layouts[type(self.element).__name__]
        for i, column_tuple in enumerate(layout):
            new_frame = tkinter.ttk.Frame(master, padding=(3, 3, 12, 12))
            new_frame.grid(row=0, column=i, sticky="nsew")
            row_no = 0
            for row_frame_key in layout[i]:
                for ui_element in self.frame_dict[row_frame_key]:
                    row_no += ui_element.body(new_frame, row_no, left=True)
        if "fastq" in self.frame_dict:
            if self.element.counts_file is not None:
                self.toggle.rb_counts.invoke()
            else:
                self.toggle.rb_fastq.invoke()

    def validate(self):
        """
        Called when the user chooses "OK", before closing the box.

        Also checks that child name is unique.
        """
        for tk_list in list(self.frame_dict.values()):
            if not all(x.validate() for x in tk_list):
                return False

        if self.element.parent is not None:
            if self.element not in self.element.parent.children:
                if self.name_entry.value.get() in self.element.parent.child_names():
                    tkinter.messagebox.showwarning("", "Sibling names must be unique.")
                    return False

        return True

    def apply(self):
        """
        Called when the user chooses "OK" and the box closes.
        """
        # apply all changes to the config object
        for tk_list in list(self.frame_dict.values()):
            for tk_element in tk_list:
                tk_element.apply()

        # use the configuration dictionary to change values
        if isinstance(self.element, SeqLib):
            self.element.configure(clear_nones(self.element_cfg))
        else:
            self.element.configure(
                clear_nones(self.element_cfg), configure_children=False
            )

        # insert into the object if necessary
        if self.element.parent is not None:
            if self.element not in self.element.parent.children:
                self.element.parent.add_child(self.element)




import os
import logging
import pandas as pd
import collections
import getpass
import time
from typing import Optional


#: Dictionary specifying available scoring methods for the analysis
#: Key is the internal name of the method, value is the GUI label
#: For command line options, internal name is used for the option string itself
#: and the value is the help string
SCORING_METHODS = collections.OrderedDict(
    [
        ("WLS", "weighted least squares"),
        ("ratios", "log ratios (Enrich2)"),
        ("counts", "counts only"),
        ("OLS", "ordinary least squares"),
        ("simple", "log ratios (old Enrich)"),
    ]
)


#: Dictionary specifying available scoring methods for the analysis
#: Key is the internal name of the method, value is the GUI label
#: For command line options, internal name is used for the option string itself
#: and the value is the help string
LOGR_METHODS = collections.OrderedDict(
    [
        ("wt", "wild type"),
        ("complete", "library size (complete cases)"),
        ("full", "library size (all reads)"),
    ]
)


#: List specifying valid labels in their sorted order
#: Sorted order is the order in which they should be calculated in
ELEMENT_LABELS = ["barcodes", "identifiers", "variants", "synonymous"]


def fix_filename(s):
    """
    Clean up a file name by removing invalid characters and converting 
    spaces to underscores.

    :param str s: file name
    :return: cleaned file name
    :rtype: str
    """
    fname = "".join(c for c in s if c.isalnum() or c in (" ._~"))
    fname = fname.replace(" ", "_")
    return fname


class StoreManager(object):
    """
    Abstract class for all data-containing classes
    (:py:class:`~enrich2.seqlib.seqlib.SeqLib`,
    :py:class:`~enrich2.selection.selection.Selection`, and
    :py:class:`~enrich2.experiment.experiment.Experiment`).

    Contains common operations for all data-containing classes, such as HDF5
    store and directory management.
    """

    store_suffix = Optional[str]
    has_store = True
    treeview_class_name = Optional[str]

    def __init__(self):
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

        # general data members
        self._name = None
        self.name = "Unnamed" + self.__class__.__name__
        self.parent = None
        self._labels = list()

        # metadata members
        self.username = getpass.getuser()
        self.creationtime = time.asctime()

        # HDF5 data
        self.store_cfg = None
        self.store_path = None
        self.store = None
        self.chunksize = 100000

        # output locations
        self._output_dir = None
        self._output_dir_override = None
        self._plot_dir = None
        self._tsv_dir = None

        # analysis parameters
        self._scoring_method = None
        self._logr_method = None

        self._force_recalculate = None
        self._component_outliers = None
        self._plots_requested = None
        self._tsv_requested = None

        # GUI variables
        self.treeview_id = None
        self.treeview_info = None

        # Plotting options
        self.plot_options = None

    def child_labels(self):
        """
        Returns a list of labels shared by every child.
        """
        shared = list()
        for x in self.children:
            shared.extend(x.labels)
        shared = collections.Counter(shared)
        shared = [x for x in list(shared.keys()) if shared[x] == len(self.children)]
        return sorted(shared, key=lambda a: ELEMENT_LABELS.index(a))

    @property
    def labels(self):
        """
        Property for returning labels shared by all children.

        If the object has no children (such as a SeqLib), the object's _labels
        are returned.
        """
        if self.children is None:
            return self._labels
        else:
            if len(self.children) > 0:
                return self.child_labels()
            else:
                return self._labels

    @property
    def force_recalculate(self):
        """
        This property should only be set for the root element. All other
        elements in the analysis should have ``None``.

        Recursively traverses up the config tree to find the root element.
        """
        if self._force_recalculate is None:
            if self.parent is not None:
                return self.parent.force_recalculate
            else:
                raise ValueError(
                    "Forced recalculation option not specified "
                    "at root [{}]".format(self.name)
                )
        else:
            return self._force_recalculate

    @force_recalculate.setter
    def force_recalculate(self, value):
        """
        Make sure the *value* is valid and set it.
        """
        if value in (True, False):
            self._force_recalculate = value
        else:
            raise ValueError(
                "Invalid setting '{}' for force_recalculate "
                "[{}]".format(value, self.name)
            )

    @property
    def component_outliers(self):
        """
        This property should only be set for the root element. All other
        elements in the analysis should have ``None``.

        Recursively traverses up the config tree to find the root element.
        """
        if self._component_outliers is None:
            if self.parent is not None:
                return self.parent.component_outliers
            else:
                raise ValueError(
                    "Calculate component outliers option not "
                    "specified at root [{}]".format(self.name)
                )
        else:
            return self._component_outliers

    @component_outliers.setter
    def component_outliers(self, value):
        """
        Make sure the *value* is valid and set it.
        """
        if value in (True, False):
            self._component_outliers = value
        else:
            raise ValueError(
                "Invalid setting '{}' for component_outliers "
                "[{}]".format(value, self.name)
            )

    @property
    def plots_requested(self):
        """
        This property should only be set for the root element. All other
        elements in the analysis should have ``None``.

        Recursively traverses up the config tree to find the root element.
        """
        if self._plots_requested is None:
            if self.parent is not None:
                return self.parent.plots_requested
            else:
                raise ValueError(
                    "Make plots option not specified at root " "[{}]".format(self.name)
                )
        else:
            return self._plots_requested

    @plots_requested.setter
    def plots_requested(self, value):
        """
        Make sure the *value* is valid and set it.
        """
        if value in (True, False):
            self._plots_requested = value
        else:
            raise ValueError(
                "Invalid setting '{}' for plots_requested "
                "[{}]".format(value, self.name)
            )

    @property
    def tsv_requested(self):
        """
        This property should only be set for the root element. All other
        elements in the analysis should have ``None``.

        Recursively traverses up the config tree to find the root element.
        """
        if self._tsv_requested is None:
            if self.parent is not None:
                return self.parent.tsv_requested
            else:
                raise ValueError(
                    "Write tsv option not specified at root " "[{}]".format(self.name)
                )
        else:
            return self._tsv_requested

    @tsv_requested.setter
    def tsv_requested(self, value):
        """
        Make sure the *value* is valid and set it.
        """
        if value in (True, False):
            self._tsv_requested = value
        else:
            raise ValueError(
                "Invalid setting '{}' for tsv_requested [{}]".format(value, self.name)
            )

    @property
    def scoring_method(self):
        """
        This property should only be set for the root element. All other
        elements in the analysis should have ``None``.

        Recursively traverses up the config tree to find the root element.
        """
        if self._scoring_method is None:
            if self.parent is not None:
                return self.parent.scoring_method
            else:
                raise ValueError(
                    "Scoring method not specified at root [{}]".format(self.name)
                )
        else:
            return self._scoring_method

    @scoring_method.setter
    def scoring_method(self, value):
        """
        Make sure the *value* is valid and set it.
        """
        if value in list(SCORING_METHODS.keys()):
            self._scoring_method = value
        else:
            raise ValueError(
                "Invalid setting for scoring_method [{}]".format(self.name)
            )

    @property
    def output_dir(self):
        """
        Elements that have ``None`` as their output directory value (*i.e.*
        no value set in the confic object) will automatically use the output
        directory from the parent.
        """
        if self._output_dir is None:
            if self.parent is not None:
                return self.parent.output_dir
            else:
                raise ValueError(
                    "No output directory specified at top level "
                    "[{}]".format(self.name)
                )
        else:
            return self._output_dir

    @output_dir.setter
    def output_dir(self, dirname):
        """
        Sets the object's base output directory to *dirname* and creates the
        directory if it doesn't exist.
        """
        try:
            dirname = os.path.expanduser(dirname)  # handle leading '~'
        except AttributeError as e:
            raise AttributeError("Invalid input for output directory: {}".format(e))
        try:
            if not os.path.exists(dirname):
                os.makedirs(dirname)
        except OSError as e:
            raise OSError("Failed to create output directory: {}".format(e))
        self._output_dir = dirname

    @property
    def output_dir_override(self):
        """
        This property should only be set for the root element. All other
        elements in the analysis should have ``None``.

        Recursively traverses up the config tree to find the root element.
        """
        if self._output_dir_override is None:
            if self.parent is not None:
                return self.parent.output_dir_override
            else:
                raise ValueError(
                    "Output directory override not specified at root [{}]".format(
                        self.name
                    )
                )
        else:
            return self._output_dir_override

    @output_dir_override.setter
    def output_dir_override(self, value):
        """
        Make sure the *value* is valid and set it.
        """
        if value in (True, False):
            self._output_dir_override = value
        else:
            raise ValueError(
                "Invalid setting '{}' for output_dir_override [{}]".format(
                    value, self.name
                )
            )

    @property
    def plot_dir(self):
        """
        Creates the plot directory when first requested if it doesn't exist.

        The plot directory is ``<output directory>/plots/<object name>/``.
        """
        if self._plot_dir is None:
            dirname = os.path.join(
                self.output_dir,
                "plots",
                "{}_{}".format(fix_filename(self.name), self.store_suffix),
            )
            try:
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
            except OSError as e:
                raise OSError("Failed to create plotting directory: {}".format(e))
            self._plot_dir = dirname
        return self._plot_dir

    @property
    def tsv_dir(self):
        """
        Creates the tsv directory when first requested if it doesn't exist.

        The plot directory is ``<output directory>/tsv/<object name>/``.
        """
        if self._tsv_dir is None:
            dirname = os.path.join(
                self.output_dir,
                "tsv",
                "{}_{}".format(fix_filename(self.name), self.store_suffix),
            )
            try:
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
            except OSError as e:
                raise OSError("Failed to create tsv directory: {}".format(e))
            self._tsv_dir = dirname
        return self._tsv_dir

    @property
    def logr_method(self):
        """
        This property should only be set for the root element. All other
        elements in the analysis should have ``None``.

        Recursively traverses up the config tree to find the root element.
        """
        if self._logr_method is None:
            if self.parent is not None:
                return self.parent.logr_method
            else:
                raise AttributeError(
                    "Log-ratio method not specified for root "
                    "element [{}]".format(self.name)
                )
        else:
            return self._logr_method

    @logr_method.setter
    def logr_method(self, value):
        """
        Make sure the *value* is valid and set it.
        """
        if value in list(LOGR_METHODS.keys()):
            self._logr_method = value
        else:
            raise ValueError(
                "Invalid setting '{}' for log-ratio method "
                "[{}]".format(value, self.name)
            )

    @property
    def children(self):
        """
        Property for returning all children of this element.

        Calls the objects :py:meth:`_children` method. This allows the
        property to be overloaded without creating a new property in the
        derived classes.
        """
        return self._children()

    def _children(self):
        """
        Pure virtual method for returning children.
        """
        raise NotImplementedError("must be implemented by subclass")

    def remove_child_id(self, tree_id):
        """
        Pure virtual method for removing a child with Treeview id *tree_id*
        from this element's children.
        """
        raise NotImplementedError("must be implemented by subclass")

    def add_child(self, child):
        """
        Pure virtual method for adding a child.
        """
        raise NotImplementedError("must be implemented by subclass")

    def child_names(self):
        """
        Return a list of the ``name`` attributes for all children.
        """
        try:
            names = [x.name for x in self.children]
        except AttributeError:
            raise AttributeError("No name set for child [{}]".format(self.name))
        else:
            return names

    def add_label(self, x):
        """
        Add element label to this object.
        """
        labels = set(self._labels)
        if isinstance(x, str):
            if x in ELEMENT_LABELS:
                labels.update([x])
            else:
                raise ValueError("Invalid element label '{}' [{}]".format(x, self.name))
        else:
            raise AttributeError("Failed to add labels [{}]".format(self.name))
        # sort based on specified order
        self._labels = sorted(labels, key=lambda a: ELEMENT_LABELS.index(a))

    def configure(self, cfg):
        """
        Set up the object using the config object *cfg*, usually derived from
        a ``.json`` file.
        """
        try:
            self.name = cfg["name"]
            if "output directory" in cfg:
                if self.output_dir_override:
                    self.logger.warning(
                        "Using command line supplied output "
                        "directory instead of config file output "
                        "directory"
                    )
                else:
                    self.output_dir = cfg["output directory"]
            if "store" in cfg:
                self.store_cfg = True
                if not os.path.exists(cfg["store"]):
                    raise IOError(
                        'Specified store file "{}" not found'.format(cfg["store"]),
                        self.name,
                    )
                elif os.path.splitext(cfg["store"])[-1].lower() != ".h5":
                    raise ValueError(
                        "Unrecognized store file extension for "
                        '"{}"'.format(cfg["store"]),
                        self.name,
                    )
                else:
                    self.store_path = cfg["store"]
                    self.logger.info(
                        'Using specified HDF5 data store "{}"'.format(self.store_path)
                    )
            else:
                self.store_cfg = False
                self.store_path = None
        except KeyError as key:
            raise KeyError("Missing required config value {}".format(key), self.name)

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for
        dumping to a config file.
        """
        cfg = {"name": self.name}
        if self.store_cfg:
            cfg["store"] = self.store_path
        if self.output_dir is not None:
            if self.parent is not None:
                if self.output_dir != self.parent.output_dir:
                    cfg["output directory"] = self.output_dir
            else:
                cfg["output directory"] = self.output_dir
        return cfg

    def validate(self):
        """
        Pure virtual method for making sure configured object is valid.
        """
        raise NotImplementedError("must be implemented by subclass")

    def store_open(self, children=False, force_delete=True):
        """
        Open the HDF5 file associated with this object. If the
        ``force_recalculate`` option is selected and ``force_delete`` is
        ``True``, the existing tables under ``'/main'`` will be deleted upon
        opening.

        This method needs a lot more error checking.
        """
        if self.has_store:
            if not self.store_cfg:
                self.store_path = os.path.join(
                    self.output_dir,
                    "{}_{}.h5".format(fix_filename(self.name), self.store_suffix),
                )
                if os.path.exists(self.store_path):
                    self.logger.info(
                        'Found existing HDF5 data store "{}"'.format(self.store_path)
                    )
                else:
                    self.logger.info(
                        'Creating new HDF5 data store "{}"'.format(self.store_path)
                    )
            self.store = pd.HDFStore(self.store_path)
            if self.force_recalculate and force_delete:
                if "/main" in self.store:
                    self.logger.info("Deleting existing calculated values")
                    self.store.remove("/main")
                else:
                    self.logger.warning("No existing calculated values in file")

        if children and self.children is not None:
            for child in self.children:
                child.store_open(children=True)

    def store_close(self, children=False):
        # needs more error checking
        if self.has_store:
            self.store.close()
            self.store = None
        if children and self.children is not None:
            for child in self.children:
                child.store_close(children=True)

    def get_metadata(self, key, store=None):
        """
        Retrieve the Enrich2 metadata dictionary from the HDF5 store.

        *key* is the name of the group or node in the HDF5 data store.

        Returns the metadata dictionary for *key*. If no metadata has been set
        for *key*, returns ``None``.

        *store* can be an external open HDFStore (used when copying metadata
        from raw counts). If it is ``None``, use this object's store.

        """
        if store is None:
            store = self.store
        try:
            metadata = store.get_storer(key).attrs["enrich2"]
        except AttributeError:
            if store is self.store:  # store parameter was None
                raise AttributeError(
                    "Invalid HDF store node '{}' [{}]".format(key, self.name)
                )
            else:
                raise AttributeError(
                    "Invalid external HDF store node '{}' in "
                    "'{}' [{}]".format(key, store.filename, self.name)
                )
        except KeyError:  # no enrich2 metadata
            return None
        else:
            return metadata

    def set_metadata(self, key, d, update=True):
        """
        Replace or update the metadata dictionary from the HDF5 store.

        *key* is the name of the group or node in the HDF5 data store.

        *d* is the dictionary containing the new metadata.

        If *update* is ``False``, *d* completely replaces the existing metadata
        for *key*. Otherwise, *d* updates the existing metadata using standard
        dictionary update.
        """
        # get existing for update
        # also performs check for the node, so we don't check here
        metadata = self.get_metadata(key)
        if metadata is None or not update:
            metadata = d
        else:
            metadata.update(d)
        self.store.get_storer(key).attrs["enrich2"] = metadata

    def calculate(self):
        """
        Pure virtual method that defines how the data are calculated.
        """
        raise NotImplementedError("must be implemented by subclass")

    def make_plots(self):
        """
        Pure virtual method for creating plots.
        """
        raise NotImplementedError("must be implemented by subclass")

    def write_tsv(self):
        """
        Pure virtual method for writing tsv files.
        """
        raise NotImplementedError("must be implemented by subclass")

    def write_table_tsv(self, key):
        """
        Write the table under *key* as a tsv file.

        Files are written to a ``tsv`` directory in the default output
        location.

        File names are the HDF5 key with ``'_'`` substituted for ``'/'``.
        """
        fname = key.strip("/")  # remove leading slash
        fname = fname.replace("/", "_") + ".tsv"
        self.store[key].to_csv(os.path.join(self.tsv_dir, fname), sep="\t", na_rep="NA")

    def check_store(self, key):
        """
        Checks to see if a particular data frame in the HDF5 store already
        exists.

        Args:
            key (str): key for the requested data frame

        Returns:
            bool: True if the key exists in the HDF5 store, else False.
        """
        if key in list(self.store.keys()):
            self.logger.info("Found existing '{}'".format(key))
            return True
        else:
            return False

    def map_table(
        self,
        source,
        destination,
        source_query=None,
        row_callback=None,
        row_callback_args=None,
        destination_data_columns=None,
    ):
        """
        Converts source table into destination table.

        This method really needs a better name.
        """
        if destination in list(self.store.keys()):
            # remove the current destination table because we are using append
            # append takes the "min_itemsize" argument, and put doesn't
            self.logger.info("Overwriting existing '{}'".format(destination))
            self.store.remove(destination)

        # turn the single table name into a list to use select_as_multiple
        if isinstance(source, str):
            source = [source]

        # assumes the source tables all have the same index
        # find the min_itemsize
        max_index_length = self.store.select_column(source[0], "index").map(len).max()
        for df in self.store.select_as_multiple(
            source, source_query, chunksize=self.chunksize, selector=source[0]
        ):
            if row_callback is not None:
                df = df.apply(row_callback, args=row_callback_args, axis="columns")
            if destination not in self.store:
                if destination_data_columns is None:
                    # if not specified, index all columns
                    destination_data_columns = list(df.columns)
                self.store.append(
                    destination,
                    df,
                    min_itemsize={"index": max_index_length},
                    data_columns=destination_data_columns,
                )
            else:
                self.store.append(destination, df)

    def combined_index(self, tables):
        """
        Return an index containing all elements in *tables*
        """
        shared = pd.Index()
        for t in tables:
            shared = shared.union(pd.Index(self.store.select_column(t, "index")))
        return shared

    def get_root(self):
        if self.parent is not None:
            return self.parent.get_root()
        else:
            return self




import logging
import pandas as pd
import numpy as np
import scipy.stats as stats
from matplotlib.backends.backend_pdf import PdfPages
import os.path
from .storemanager import StoreManager
from .condition import Condition
from .constants import WILD_TYPE_VARIANT
from .sfmap import sfmap_plot
from .dataframe import singleton_dataframe
from .random_effects import rml_estimator


class Experiment(StoreManager):
    """
    Class for a coordinating multiple :py:class:`~.selection.Selection`
    objects. Creating an
    :py:class:`~experiment.Experiment` requires a valid *config* object,
    usually from a ``.json`` configuration file.
    """

    store_suffix = "exp"
    treeview_class_name = "Experiment"

    def __init__(self):
        StoreManager.__init__(self)
        self.conditions = list()
        self._wt = None
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

    @property
    def wt(self):
        if self.has_wt_sequence():
            if self._wt is None:
                self._wt = self.selection_list()[0].wt.duplicate(self.name)
            return self._wt
        else:
            if self._wt is not None:
                raise ValueError(
                    "Experiment should not contain wild type "
                    "sequence [{}]".format(self.name)
                )
            else:
                return None

    def configure(self, cfg, configure_children=True):
        """
        Set up the :py:class:`~experiment.Experiment` using the *cfg* object,
        usually from a ``.json`` configuration file.
        """
        StoreManager.configure(self, cfg)
        self.logger = logging.getLogger(
            "{}.{} - {}".format(__name__, self.__class__.__name__, self.name)
        )
        if configure_children:
            if "conditions" not in cfg:
                raise KeyError(
                    "Missing required config value {} [{}]"
                    "".format("conditions", self.name)
                )
            for cnd_cfg in cfg["conditions"]:
                cnd = Condition()
                cnd.configure(cnd_cfg)
                self.add_child(cnd)

        selection_names = [x.name for x in self.selection_list()]
        if len(set(selection_names)) != len(selection_names):
            raise ValueError("Non-unique selection names [{}]" "".format(self.name))

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for
        dumping to a config file.
        """
        cfg = StoreManager.serialize(self)
        cfg["conditions"] = [child.serialize() for child in self.children]
        return cfg

    def _children(self):
        """
        Method bound to the ``children`` property. Returns a list of all
        :py:class:`~condition.Condition` objects belonging to this object,
        sorted by name.
        """
        return sorted(self.conditions, key=lambda x: x.name)

    def add_child(self, child):
        """
        Add a selection.
        """
        if child.name in self.child_names():
            raise ValueError(
                "Non-unique condition name '{}' [{}]" "".format(child.name, self.name)
            )
        child.parent = self
        self.conditions.append(child)

    def remove_child_id(self, tree_id):
        """
        Remove the reference to a :py:class:`~condition.Condition` with
        Treeview id *tree_id*.
        """
        self.conditions = [x for x in self.conditions if x.treeview_id != tree_id]

    def selection_list(self):
        """
        Return the :py:class:`~selection.Selection` objects as a list.
        """
        selections = list()
        for cnd in self.children:
            selections.extend(cnd.children)
        return selections

    def validate(self):
        """
        Calls validate on all child Conditions. Also checks the wild type
        sequence status.
        """
        # check the wild type sequences
        if self.has_wt_sequence():
            for child in self.selection_list()[1:]:
                if self.selection_list()[0].wt != child.wt:
                    self.logger.warning("Inconsistent wild type sequences")
                    break

        for child in self.children:
            child.validate()

    def is_coding(self):
        """
        Return ``True`` if the all :py:class:`~selection.Selection` in the
        :py:class:`~experiment.Experiment` count protein-coding variants, else
        ``False``.
        """
        return all(x.is_coding() for x in self.selection_list())

    def has_wt_sequence(self):
        """
        Return ``True`` if the all :py:class:`~selection.Selection` in the
        :py:class:`~experiment.Experiment` have a wild type sequence, else
        ``False``.
        """
        return all(x.has_wt_sequence() for x in self.selection_list())

    def calculate(self):
        """
        Calculate scores for all :py:class:`~selection.Selection` objects.
        """
        if len(self.labels) == 0:
            raise ValueError(
                "No data present across all conditions [{}]" "".format(self.name)
            )
        for s in self.selection_list():
            s.calculate()
        self.combine_barcode_maps()
        for label in self.labels:
            self.calc_counts(label)
            if self.scoring_method != "counts":
                self.calc_shared_full(label)
                self.calc_shared(label)
                self.calc_scores(label)
                if label != "barcodes":
                    self.calc_pvalues_wt(label)

    def combine_barcode_maps(self):
        """
        Combine all barcode maps for :py:class:`~selection.Selection` objects
        into a single data frame and store it in ``'/main/barcodemap'``.

        If multiple variants or IDs map to the same barcode, only the first one
        will be present in the barcode map table.

        The ``'/main/barcodemap'`` table is not created if no
        :py:class:`~selection.Selection` has barcode map information.
        """
        if self.check_store("/main/barcodemap"):
            return

        bcm = None
        for sel in self.selection_list():
            if "/main/barcodemap" in list(sel.store.keys()):
                if bcm is None:
                    bcm = sel.store["/main/barcodemap"]
                else:
                    bcm = bcm.join(
                        sel.store["/main/barcodemap"], rsuffix=".drop", how="outer"
                    )
                    new = bcm.loc[pd.isnull(bcm)["value"]]
                    bcm.loc[new.index, "value"] = new["value.drop"]
                    bcm.drop("value.drop", axis="columns", inplace=True)
        if bcm is not None:
            bcm.sort_values("value", inplace=True)
            self.store.put(
                "/main/barcodemap", bcm, format="table", data_columns=bcm.columns
            )

    def calc_counts(self, label):
        """
        Create a data frame of all counts in this Experiment. This data frame
        is not used for any calculations, but is provided to facilitate
        exploration of the data set.
        """
        if self.check_store("/main/{}/counts".format(label)):
            return

        # create columns multi-index
        # has to be lex-sorted for multi-slicing to work
        self.logger.info("Creating column multi-index for counts ({})".format(label))
        conditions_index = list()
        selections_index = list()
        values_index = list()

        for cnd in self.children:
            for sel in cnd.children:
                conditions_index.extend([cnd.name] * len(sel.timepoints))
                selections_index.extend([sel.name] * len(sel.timepoints))
                values_index.extend(["c_{}".format(x) for x in sorted(sel.timepoints)])
        columns = pd.MultiIndex.from_tuples(
            list(zip(conditions_index, selections_index, values_index)),
            names=["condition", "selection", "timepoint"],
        )

        # create union index
        self.logger.info("Creating row index for counts ({})".format(label))
        combined = None
        first = True
        for s in self.selection_list():
            if first:
                combined = s.store.select(
                    "/main/{}/counts_unfiltered" "".format(label), "columns='index'"
                ).index
                first = False
            else:
                combined = combined.join(
                    s.store.select(
                        "/main/{}/counts_unfiltered".format(label), "columns='index'"
                    ).index,
                    how="outer",
                )

        # create and fill the data frames
        self.logger.info(
            "Populating Experiment data frame with counts ({})".format(label)
        )
        data = pd.DataFrame(index=combined, columns=columns)
        for cnd in self.children:
            for sel in cnd.children:
                sel_data = sel.store.select(
                    "/main/{}/counts_unfiltered" "".format(label)
                )
                for tp in sel.timepoints:
                    data[(cnd.name, sel.name, "c_{}".format(tp))] = sel_data[
                        "c_{}".format(tp)
                    ]

        self.store.put("/main/{}/counts".format(label), data, format="table")

    def calc_shared_full(self, label):
        """
        Use joins to create a data frame containing all scores across all
        Selections in the Experiment.
        """
        if self.check_store("/main/{}/scores_shared_full".format(label)):
            return

        # create columns multi-index
        # has to be lex-sorted for multi-slicing to work
        self.logger.info("Creating column multi-index for scores ({})")
        conditions_index = list()
        selections_index = list()
        values_index = list()

        if self.scoring_method == "simple":
            values_list = ["score"]
        else:
            values_list = ["score", "SE"]

        for cnd in self.children:
            for sel in cnd.children:
                conditions_index.extend([cnd.name] * len(values_list))
                selections_index.extend([sel.name] * len(values_list))
                values_index.extend(sorted(values_list))
        columns = pd.MultiIndex.from_tuples(
            list(zip(conditions_index, selections_index, values_index)),
            names=["condition", "selection", "value"],
        )

        # create union index
        self.logger.info("Creating row index for scores ({})".format(label))
        combined = None
        first = True
        for s in self.selection_list():
            if first:
                combined = s.store.select(
                    "/main/{}/scores".format(label), "columns='index'"
                ).index
                first = False
            else:
                combined = combined.join(
                    s.store.select(
                        "/main/{}/scores".format(label), "columns='index'"
                    ).index,
                    how="outer",
                )

        # create and fill the data frames
        self.logger.info(
            "Populating Experiment data frame with scores ({})".format(label)
        )
        data = pd.DataFrame(index=combined, columns=columns).astype(float)
        for cnd in self.children:
            for sel in cnd.children:
                sel_data = sel.store.select("/main/{}/scores".format(label))
                for v in values_list:
                    data.loc[:, (cnd.name, sel.name, v)] = sel_data[v]
        self.store.put(
            "/main/{}/scores_shared_full".format(label), data, format="table"
        )

    def calc_shared(self, label):
        """
        Get the subset of scores that are shared across all Selections in each
        Condition.
        """
        if self.check_store("/main/{}/scores_shared".format(label)):
            return

        idx = pd.IndexSlice

        self.logger.info(
            "Identifying subset shared across all Selections ({})".format(label)
        )
        data = self.store.select("/main/{}/scores_shared_full".format(label))

        # identify variants found in all selections in at least one condition
        complete = np.full(len(data.index), False, dtype=bool)
        for cnd in data.columns.levels[0]:
            complete = np.logical_or(
                complete, data.loc[:, idx[cnd, :, :]].notnull().all(axis="columns")
            )

        data = data.loc[complete]

        self.store.put("/main/{}/scores_shared".format(label), data, format="table")

    def calc_scores(self, label):
        """
        Combine the scores and standard errors within each condition.
        """
        if self.check_store("/main/{}/scores".format(label)):
            return

        self.logger.info("Calculating per-condition scores ({})".format(label))

        # set up new data frame
        shared_index = self.store.select(
            "/main/{}/scores_shared" "".format(label), columns="index"
        ).index
        columns = pd.MultiIndex.from_product(
            [sorted(self.child_names()), sorted(["score", "SE", "epsilon"])],
            names=["condition", "value"],
        )
        data = pd.DataFrame(np.nan, index=shared_index, columns=columns)
        del shared_index
        del columns

        # set up local variables
        idx = pd.IndexSlice

        score_df = self.store.select("/main/{}/scores_shared".format(label))
        if self.scoring_method == "simple":
            # special case for simple ratios that have no SE
            # calculates the average score
            for cnd in score_df.columns.levels[0]:
                data.loc[:, idx[cnd, "score"]] = score_df.loc[
                    :, idx[cnd, :, "score"]
                ].mean(axis=1)
        else:
            for cnd in score_df.columns.levels[0]:
                y = np.array(score_df.loc[:, idx[cnd, :, "score"]].values).T
                sigma2i = np.array(score_df.loc[:, idx[cnd, :, "SE"]].values ** 2).T

                # single replicate of the condition
                if y.shape[0] == 1:
                    data.loc[:, idx[cnd, "score"]] = y.ravel()
                    data.loc[:, idx[cnd, "SE"]] = np.sqrt(sigma2i).ravel()
                    data.loc[:, idx[cnd, "epsilon"]] = 0.0
                # multiple replicates
                else:
                    betaML, var_betaML, eps = rml_estimator(y, sigma2i)
                    data.loc[:, idx[cnd, "score"]] = betaML
                    data.loc[:, idx[cnd, "SE"]] = np.sqrt(var_betaML)
                    data.loc[:, idx[cnd, "epsilon"]] = eps

                # special case for normalized wild type variant
                if self.logr_method == "wt" and WILD_TYPE_VARIANT in data.index:
                    data.loc[WILD_TYPE_VARIANT, idx[:, "SE"]] = 0.0
                    data.loc[WILD_TYPE_VARIANT, idx[:, "score"]] = 0.0
                    data.loc[WILD_TYPE_VARIANT, idx[:, "epsilon"]] = 0.0

        # store the data
        self.store.put("/main/{}/scores".format(label), data, format="table")

    def calc_pvalues_wt(self, label):
        """
        Calculate uncorrected pvalue for each variant compared to wild type.
        """
        if self.check_store("/main/{}/scores_pvalues_wt".format(label)):
            return

        idx = pd.IndexSlice

        wt = self.store.select(
            "/main/{}/scores".format(label), "index=WILD_TYPE_VARIANT"
        )
        if len(wt) == 0:  # no wild type score
            self.logger.info(
                "Failed to find wild type score, skipping wild type p-value calculations"
            )
            return
        data = self.store.select(
            "/main/{}/scores".format(label), "index!=WILD_TYPE_VARIANT"
        )

        columns = pd.MultiIndex.from_product(
            [sorted(self.child_names()), sorted(["z", "pvalue_raw"])],
            names=["condition", "value"],
        )
        result_df = pd.DataFrame(index=data.index, columns=columns, dtype=float)

        condition_labels = data.columns.levels[0]
        for cnd in condition_labels:
            result_df.loc[:, idx[cnd, "z"]] = np.absolute(
                wt.loc[WILD_TYPE_VARIANT, idx[cnd, "score"]]
                - data.loc[:, idx[cnd, "score"]]
            ) / np.sqrt(
                wt.loc[WILD_TYPE_VARIANT, idx[cnd, "SE"]] ** 2
                + data.loc[:, idx[cnd, "SE"]] ** 2
            )
            result_df.loc[:, idx[cnd, "pvalue_raw"]] = 2 * stats.norm.sf(
                result_df.loc[:, idx[cnd, "z"]]
            )

        self.store.put(
            "/main/{}/scores_pvalues_wt".format(label), result_df, format="table"
        )

    def calc_pvalues_pairwise(self, label):
        """
        Calculate pvalues for each variant in each pair of Conditions.
        """
        if self.check_store("/main/{}/scores_pvalues".format(label)):
            return

        data = self.store["/main/{}/scores".format(label)]

        cnd1_index = list()
        cnd2_index = list()
        values_index = list()

        values_list = ["z", "pvalue_raw"]

        condition_labels = data.columns.levels[0]
        for i, cnd1 in enumerate(condition_labels):
            for cnd2 in condition_labels[i + 1 :]:
                cnd1_index.extend([cnd1] * len(values_list))
                cnd2_index.extend([cnd2] * len(values_list))
                values_index.extend(sorted(values_list))
        columns = pd.MultiIndex.from_tuples(
            list(zip(cnd1_index, cnd2_index, values_index)),
            names=["condition1", "condition2", "value"],
        )

        idx = pd.IndexSlice
        result_df = pd.DataFrame(np.nan, index=data.index, columns=columns)
        for i, cnd1 in enumerate(condition_labels):
            for cnd2 in condition_labels[i + 1 :]:
                result_df.loc[:, idx[cnd1, cnd2, "z"]] = np.absolute(
                    data.loc[:, idx[cnd1, "score"]] - data.loc[:, idx[cnd2, "score"]]
                ) / np.sqrt(
                    data.loc[:, idx[cnd1, "SE"]] ** 2
                    + data.loc[:, idx[cnd2, "SE"]] ** 2
                )
                result_df.loc[:, idx[cnd1, cnd2, "pvalue_raw"]] = 2 * stats.norm.sf(
                    result_df.loc[:, idx[cnd1, cnd2, "z"]]
                )

        self.store.put(
            "/main/{}/scores_pvalues".format(label), result_df, format="table"
        )

    def make_plots(self):
        if self.plots_requested:
            self.logger.info("Creating plots")

            # sequence-function maps
            if self.scoring_method != "counts":
                if "synonymous" in self.labels:
                    pdf = PdfPages(
                        os.path.join(self.plot_dir, "sequence_function_map_aa.pdf")
                    )
                    for condition in self.children:
                        self.sfmap_wrapper(
                            condition=condition.name, pdf=pdf, coding=True
                        )
                    pdf.close()
                if "variants" in self.labels:
                    pdf = PdfPages(
                        os.path.join(self.plot_dir, "sequence_function_map_nt.pdf")
                    )
                    for condition in self.children:
                        self.sfmap_wrapper(
                            condition=condition.name, pdf=pdf, coding=False
                        )
                    pdf.close()

        for s in self.selection_list():
            s.make_plots()

    def write_tsv(self):
        """
        Write each table from the store to its own tab-separated file.

        Files are written to a ``tsv`` directory in the default output
        location. File names are the HDF5 key with ``'_'`` substituted for
        ``'/'``.
        """
        if self.tsv_requested:
            self.logger.info("Generating tab-separated output files")
            for k in list(self.store.keys()):
                self.write_table_tsv(k)
        for s in self.selection_list():
            s.write_tsv()

    def sfmap_wrapper(self, condition, pdf, coding):
        """
        Create a sequence function map for scores in *condition*.

        Uses :py:func:`~sfmap.sfmap_plot` for the plotting.
        """
        plot_options = self.get_root().plot_options

        if coding:
            label = "amino acid"
        else:
            label = "nucleotide"

        self.logger.info(
            "Creating sequence-function map ({}, {})".format(condition, label)
        )

        idx = pd.IndexSlice

        if coding:
            df_name = "/main/synonymous/scores"
        else:
            df_name = "/main/variants/scores"

        if plot_options is not None:
            data, wtseq = singleton_dataframe(
                self.store[df_name][idx[condition, "score"]],
                self.wt,
                coding=coding,
                aa_list=plot_options["aa_list"],
            )
            data_se, _ = singleton_dataframe(
                self.store[df_name][idx[condition, "SE"]],
                self.wt,
                coding=coding,
                aa_list=plot_options["aa_list"],
            )
        else:
            data, wtseq = singleton_dataframe(
                self.store[df_name][idx[condition, "score"]], self.wt, coding=coding
            )
            data_se, _ = singleton_dataframe(
                self.store[df_name][idx[condition, "SE"]], self.wt, coding=coding
            )

        # format the title
        if coding:
            title = "Amino Acid"
        else:
            title = "Nucleotide"

        if self.scoring_method in ("WLS", "OLS"):
            title += " Sequence-Function Map\n{} ({} Slope)".format(
                condition, self.scoring_method
            )
        elif self.scoring_method == "ratios":
            title += " Sequence-Function Map\n{} ({})".format(
                condition, "Enrich2 Ratio"
            )
        elif self.scoring_method == "simple":
            title += " Sequence-Function Map\n{} ({})".format(
                condition, "Simplified Ratio"
            )
        else:
            raise ValueError("Invalid scoring method", self.name)

        if plot_options is not None:
            sfmap_plot(
                df=data,
                pdf=pdf,
                style="scores",
                df_se=data_se,
                dimensions="tall",
                wt=wtseq,
                title=title,
                aa_list=plot_options["aa_list"],
                aa_label_groups=plot_options["aa_label_groups"],
            )
        else:
            sfmap_plot(
                df=data,
                pdf=pdf,
                style="scores",
                df_se=data_se,
                dimensions="tall",
                wt=wtseq,
                title=title,
            )

    def correlation_plot(self, pdf, label):
        """
        Create a triangular heatmap showing the Pearson correlation coefficient
        for each pairwise comparison of replicate scores.
        """
        pass



import logging
from .seqlib import SeqLib


class IdOnlySeqLib(SeqLib):
    """
    Class for counting data with non-variant identifiers and no associated
    FASTQ_ data.
    """

    treeview_class_name = "ID-only SeqLib"

    def __init__(self):
        SeqLib.__init__(self)
        self.identifier_min_count = 0
        self.add_label("identifiers")
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

    def configure(self, cfg):
        """
        Set up the object using the config object *cfg*, usually derived from
        a ``.json`` file.
        """
        SeqLib.configure(self, cfg)
        self.logger = logging.getLogger(
            "{}.{} - {}".format(__name__, self.__class__.__name__, self.name)
        )
        try:
            if "min count" in cfg["identifiers"]:
                self.identifier_min_count = int(cfg["identifiers"]["min count"])
        except KeyError as key:
            raise KeyError(
                "Missing required config value {key} [{name}]"
                "".format(key=key, name=self.name)
            )

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for
        dumping to a config file.
        """
        cfg = SeqLib.serialize(self)

        cfg["identifiers"] = dict()
        if self.identifier_min_count > 0:
            cfg["identifiers"]["min count"] = self.identifier_min_count

        return cfg

    def calculate(self):
        """
        Get the identifier counts from the counts file.
        """
        if not self.check_store("/main/identifiers/counts"):
            if self.counts_file is not None:
                self.counts_from_file(self.counts_file)
            else:
                raise ValueError("Missing counts file [{}]".format(self.name))
            self.save_filtered_counts(
                "identifiers", "count >= self.identifier_min_count"
            )



import tkinter as tk
import tkinter.ttk
import tkinter.simpledialog


def subtree_ids(treeview, x, level=0):
    """
    Return a list of tuples containing the ids and levels for *x* and every element below it in the Treeview *treeview*.

    The level of *x* is 0, children of *x* are 1, and so forth.
    """
    id_list = list()
    id_list.append((x, level))
    for y in treeview.get_children(x):
        id_list.extend(subtree_ids(treeview, y, level + 1))
    return id_list


class DeleteDialog(tkinter.simpledialog.Dialog):
    """
    Confirmation dialog box for deleting the selected items from the Treeview.
    """

    def __init__(self, parent_window, tree, title="Confirm Deletion"):
        self.tree = tree
        self.id_tuples = list()
        for x in self.tree.treeview.selection():
            if x not in [y[0] for y in self.id_tuples]:
                self.id_tuples.extend(subtree_ids(self.tree.treeview, x))
        tkinter.simpledialog.Dialog.__init__(self, parent_window, title)

    def body(self, master):
        """
        Generates the required text listing all elements that will be deleted.

        Displays the "OK" and "Cancel" buttons.
        """
        if len(self.id_tuples) == 0:
            message_string = "No elements selected."
        elif len(self.id_tuples) == 1:
            message_string = 'Delete "{}"?'.format(
                self.tree.get_element(self.id_tuples[0][0]).name
            )
        else:
            message_string = "Delete the following items?\n"
            for x, level in self.id_tuples:
                if level == 0:
                    bullet = "    " + "\u25C6"
                else:
                    bullet = "    " * (level + 1) + "\u25C7"
                message_string += "{bullet} {name}\n".format(
                    bullet=bullet, name=self.tree.get_element(x).name
                )
        message = tkinter.ttk.Label(master, text=message_string, justify="left")
        message.grid(row=0, sticky="w")

    def buttonbox(self):
        """
        Display only one button if there's no selection. Otherwise, use the default method to display two buttons.
        """
        if len(self.id_tuples) == 0:
            box = tk.Frame(self)

            w = tk.Button(
                box, text="OK", width=10, command=self.cancel, default="active"
            )
            w.pack(side="left", padx=5, pady=5)

            self.bind("<Return>", self.cancel)

            box.pack()
        else:
            tkinter.simpledialog.Dialog.buttonbox(self)

    def apply(self):
        """
        Called when the user chooses "OK". Performs the deletion.
        """
        for tree_id, _ in self.id_tuples:
            self.tree.delete_element(tree_id)
        self.tree.refresh_treeview()



from .variant import VariantSeqLib
from fqfa import open_compressed, parse_fastq_reads, has_fastq_ext
import logging
import sys


class BasicSeqLib(VariantSeqLib):
    """
    Class for count data from sequencing libraries with a single read for
    each variant. Creating a :py:class:`BasicSeqLib` requires a valid
    *config* object, usually from a ``.json`` configuration file.
    """

    treeview_class_name = "Basic SeqLib"

    def __init__(self):
        VariantSeqLib.__init__(self)
        self.reads = None
        self.reverse_complement_reads = False
        self.trim_start = 1
        self.trim_length = sys.maxsize
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

    def configure(self, cfg):
        """
        Set up the object using the config object *cfg*, usually derived from
        a ``.json`` file.
        """
        VariantSeqLib.configure(self, cfg)
        self.logger = logging.getLogger(
            "{}.{} - {}".format(__name__, self.__class__.__name__, self.name)
        )

        # if counts are specified, copy them later
        # else handle the FASTQ config options and check the files
        if self.counts_file is None:
            self.configure_fastq(cfg)
            try:
                if not has_fastq_ext(self.reads):
                    raise IOError(
                        "FASTQ file error: unrecognized extension "
                        "[{}]".format(self.name)
                    )
            except IOError as fqerr:
                raise IOError("FASTQ file error [{}]: {}".format(self.name, fqerr))

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for
        dumping to a config file.
        """
        cfg = VariantSeqLib.serialize(self)

        cfg["fastq"] = self.serialize_fastq()

        return cfg

    def configure_fastq(self, cfg):
        """
        Set up the object's FASTQ_ file handling and filtering options.
        """
        try:
            self.reads = cfg["fastq"]["reads"]

            if "reverse" in cfg["fastq"]:
                self.reverse_complement_reads = cfg["fastq"]["reverse"]

            if "start" in cfg["fastq"]:
                self.trim_start = cfg["fastq"]["start"]

            if "length" in cfg["fastq"]:
                self.trim_length = cfg["fastq"]["length"]

            self.filters = cfg["fastq"]["filters"]
        except KeyError as key:
            raise KeyError(
                "Missing required config value {key} [{name}]"
                "".format(key=key, name=self.name)
            )

    def serialize_fastq(self):
        """
        Serialize this object's FASTQ_ file handling and filtering options.
        """
        fastq = {"filters": self.serialize_filters()}
        fastq["reads"] = self.reads

        if self.reverse_complement_reads:
            fastq["reverse"] = True
        else:
            fastq["reverse"] = False

        if self.trim_start > 1:
            fastq["start"] = self.trim_start

        if self.trim_length < sys.maxsize:
            fastq["length"] = self.trim_length

        return fastq

    def counts_from_reads(self):
        """
        Reads the forward or reverse FASTQ_ file (reverse reads are
        reverse-complemented), performs quality-based filtering, and counts
        the variants.
        """
        df_dict = dict()

        self.logger.info("Counting variants")
        max_mut_variants = 0
        with open_compressed(self.reads) as handle:
            for fq in parse_fastq_reads(handle):
                fq.trim(start=self.trim_start, end=self.trim_start + self.trim_length -1)
                if self.reverse_complement_reads:
                    fq.reverse_complement()

                if self.read_quality_filter(fq):
                    mutations = self.count_variant(fq.sequence)
                    if mutations is None:  # too many mutations
                        max_mut_variants += 1
                        if self.report_filtered:
                            self.report_filtered_variant(fq.sequence, 1)
                    else:
                        try:
                            df_dict[mutations] += 1
                        except KeyError:
                            df_dict[mutations] = 1

        self.save_counts("variants", df_dict, raw=True)
        del df_dict

        if self.aligner is not None:
            self.logger.info("Aligned {} variants".format(self.aligner.calls))
            self.aligner_cache = None
        self.logger.info(
            "Removed {} total variants with excess mutations"
            "".format(max_mut_variants)
        )
        self.save_filter_stats()

    def calculate(self):
        """
        Counts variants from counts file or FASTQ.
        """
        if not self.check_store("/main/variants/counts"):
            if not self.check_store("/raw/variants/counts"):
                if self.counts_file is not None:
                    self.counts_from_file(self.counts_file)
                else:
                    self.counts_from_reads()
            self.save_filtered_counts("variants", "count >= self.variant_min_count")
        self.count_synonymous()



"""
Module for alignment of variants to the wild type sequence.

This module is optional, and using it will dramatically increase runtime when
counting variants. It is only recommended for users who need to count
insertion and deletion variants (i.e. not coding sequences).
"""

import numpy as np

#: Default similarity matrix used by the aligner.
#: User-defined matrices must have this format.
_simple_similarity = {
    "A": {"A": 1, "C": -1, "G": -1, "T": -1, "N": 0, "X": 0},
    "C": {"A": -1, "C": 1, "G": -1, "T": -1, "N": 0, "X": 0},
    "G": {"A": -1, "C": -1, "G": 1, "T": -1, "N": 0, "X": 0},
    "T": {"A": -1, "C": -1, "G": -1, "T": 1, "N": 0, "X": 0},
    "N": {"A": 0, "C": 0, "G": 0, "T": 0, "N": 0, "X": 0},
    "X": {"A": 0, "C": 0, "G": 0, "T": 0, "N": 0, "X": 0},
    "gap": -1,
}


class Aligner(object):
    """
    Class for performing local alignment of two DNA sequences.

    This class implements `Needleman-Wunsch <http://en.wikipedia.org/wiki/
    Needleman%E2%80%93Wunsch_algorithm>`_ local alignment.

    The :py:class:`~aligner.Aligner` requires a scoring matrix when
    created. The format is a nested dictionary, with a special ``'gap'`` entry
    for the gap penalty (this value is used for both gap opening and gap
    extension).

    The ``'X'`` nucleotide is a special case for unresolvable mismatches in
    :py:class:`~overlap.OverlapSeqLib` variant data.
    """

    _MAT = 1  # match
    _INS = 2  # insertion (with respect to wild type)
    _DEL = 3  # deletion (with respect to wild type)
    _END = 4  # end of traceback

    def __init__(self, similarity=_simple_similarity):
        similarity_keys = list(similarity.keys())
        if "gap" in similarity_keys:
            similarity_keys.remove("gap")
        for key in similarity_keys:
            if not all(x in similarity[key] for x in similarity_keys) or len(
                similarity[key]
            ) != len(similarity_keys):
                raise ValueError("Asymmetrical alignment scoring matrix")

        self.similarity = similarity
        if "gap" not in self.similarity:
            raise ValueError("No gap penalty in alignment scoring matrix.")

        self.matrix = None
        self.seq1 = None
        self.seq2 = None
        self.calls = 0

    def align(self, seq1, seq2):
        """
        Aligns the two sequences, *seq1* and *seq2* and returns a list of
        tuples describing the differences between the sequences.

        The tuple format is ``(i, j, type, length)``, where ``i`` and ``j``
        are the positions in *seq1* and *seq2*, respectively, and type is one
        of ``"match"``, ``"mismatch"``, ``"insertion"``, or ``"deletion"``.
        For indels, the ``length`` value is the number of bases inserted or
        deleted with respect to *seq1* starting at ``i``.
        """
        self.matrix = np.ndarray(
            shape=(len(seq1) + 1, len(seq2) + 1),
            dtype=np.dtype([("score", int), ("trace", np.byte)]),
        )
        seq1 = seq1.upper()
        seq2 = seq2.upper()

        # build matrix of scores/traceback information
        for i in range(len(seq1) + 1):
            self.matrix[i, 0] = (self.similarity["gap"] * i, Aligner._DEL)
        for j in range(len(seq2) + 1):
            self.matrix[0, j] = (self.similarity["gap"] * j, Aligner._INS)
        for i in range(1, len(seq1) + 1):
            for j in range(1, len(seq2) + 1):
                match = (
                    self.matrix[i - 1, j - 1]["score"]
                    + self.similarity[seq1[i - 1]][seq2[j - 1]],
                    Aligner._MAT,
                )
                delete = (
                    self.matrix[i - 1, j]["score"] + self.similarity["gap"],
                    Aligner._DEL,
                )
                insert = (
                    self.matrix[i, j - 1]["score"] + self.similarity["gap"],
                    Aligner._INS,
                )
                self.matrix[i, j] = max(delete, insert, match, key=lambda x: x[0])
        self.matrix[0, 0] = (0, Aligner._END)

        # calculate alignment from the traceback
        i = len(seq1)
        j = len(seq2)
        traceback = list()
        while i > 0 or j > 0:
            if self.matrix[i, j]["trace"] == Aligner._MAT:
                if seq1[i - 1] == seq2[j - 1]:
                    traceback.append((i - 1, j - 1, "match", None))
                else:
                    traceback.append((i - 1, j - 1, "mismatch", None))
                i -= 1
                j -= 1
            elif self.matrix[i, j]["trace"] == Aligner._INS:
                traceback.append((i - 1, j - 1, "insertion", 1))
                j -= 1
            elif self.matrix[i, j]["trace"] == Aligner._DEL:
                traceback.append((i - 1, j - 1, "deletion", 1))
                i -= 1
            elif self.matrix[i, j]["trace"] == Aligner._END:
                pass
            else:
                raise RuntimeError("Invalid value in alignment traceback.")
        traceback.reverse()

        # combine indels
        indel = None
        traceback_combined = list()
        for t in traceback:
            if t[2] == "insertion" or t[2] == "deletion":
                if indel is not None:
                    if t[2] == indel[2]:
                        indel[3] += t[3]
                    else:
                        raise RuntimeError(
                            "Aligner failed to combine indels. " "Check gap penalty."
                        )
                else:
                    indel = list(t)
            else:
                if indel is not None:
                    traceback_combined.append(tuple(indel))
                    indel = None
                traceback_combined.append(t)
        if indel is not None:
            traceback_combined.append(tuple(indel))

        self.calls += 1
        return traceback_combined




import tkinter as tk
import tkinter.ttk
import tkinter.messagebox
import tkinter.filedialog
import os.path

DEFAULT_COLUMNS = 3


class SectionLabel(object):
    def __init__(self, text):
        self.text = text

    def body(self, master, row, columns=DEFAULT_COLUMNS, **kwargs):
        label = tkinter.ttk.Label(master, text=self.text)
        label.grid(row=row, column=0, columnspan=columns, sticky="w")
        return 1

    def validate(self):
        return True

    def apply(self):
        return None

    def enable(self):
        pass

    def disable(self):
        pass


class Checkbox(object):
    def __init__(self, text, cfg, key):
        self.checkbox = None
        self.enabled = True

        self.value = tk.BooleanVar()
        self.text = text
        self.cfg = cfg
        self.key = key
        try:
            if self.cfg[self.key] not in (True, False):
                self.value.set(False)
            else:
                self.value.set(self.cfg[self.key])
        except KeyError:
            self.value.set(False)  # default to False

    def body(self, master, row, columns=DEFAULT_COLUMNS, **kwargs):
        """
        Place the required elements using the grid layout method.

        Returns the number of rows taken by this element.
        """
        self.checkbox = tkinter.ttk.Checkbutton(master, text=self.text, variable=self.value)
        self.checkbox.grid(row=row, column=0, columnspan=columns, sticky="w")
        return 1

    def validate(self):
        return True

    def apply(self):
        if self.enabled:
            self.cfg[self.key] = self.value.get()
        else:
            self.cfg[self.key] = None

    def enable(self):
        self.enabled = True
        self.checkbox.state(["!disabled"])

    def disable(self):
        self.enabled = False
        self.checkbox.state(["disabled"])


class MyEntry(object):
    """
    Base class for labeled Entry fields.

    *text* is the Label/error box text.
    """

    def __init__(self, text, cfg, key, optional=False):
        self.entry = None
        self.enabled = True

        self.value = tk.StringVar()
        self.text = text
        self.cfg = cfg
        self.key = key
        self.optional = optional
        try:
            if self.cfg[self.key] is None:
                self.value.set("")
            else:
                self.value.set(self.cfg[self.key])
        except KeyError:
            self.value.set("")

    def body(self, master, row, columns=DEFAULT_COLUMNS, **kwargs):
        """
        Place the required elements using the grid layout method.

        Returns the number of rows taken by this element.
        """
        label = tkinter.ttk.Label(master, text=self.text)
        label.grid(row=row, column=0, columnspan=1, sticky="e")
        self.entry = tkinter.ttk.Entry(master, textvariable=self.value)
        self.entry.grid(row=row, column=1, columnspan=columns - 1, sticky="ew")
        return 1

    def validate(self):
        """
        Validates the input. Returns ``True`` unless the field is blank and
        *optional* is ``False``.
        """
        if not self.enabled:
            return True
        elif not self.optional and len(self.value.get()) == 0:
            tkinter.messagebox.showwarning("", "{} not specified.".format(self.text))
            return False
        else:
            return True

    def apply(self):
        if self.enabled and len(self.value.get()) > 0:
            self.cfg[self.key] = self.value.get()
        else:
            self.cfg[self.key] = None

    def enable(self):
        self.enabled = True
        self.entry.state(["!disabled"])

    def disable(self):
        self.enabled = False
        self.entry.state(["disabled"])


class FileEntry(MyEntry):
    """
    Creates a labeled Entry field for a file or directory.

    *text* is the Label/error box text.
    *directory* is ``True`` if selecting a directory (instead of a file).
    *extensions* is a list of valid file endings

    """

    def __init__(
        self, text, cfg, key, optional=False, directory=False, extensions=None
    ):
        MyEntry.__init__(self, text, cfg, key, optional)
        self.choose = None
        self.clear = None

        self.directory = directory
        if extensions is not None:
            self.extensions = [x.lower() for x in extensions]
        else:
            self.extensions = None

    def body(self, master, row, columns=DEFAULT_COLUMNS, **kwargs):
        """
        Place the required elements using the grid layout method.

        Returns the number of rows taken by this element.
        """
        label = tkinter.ttk.Label(master, text=self.text)
        label.grid(row=row, column=0, columnspan=1, sticky="e")
        self.entry = tkinter.ttk.Entry(master, textvariable=self.value)
        self.entry.grid(row=row, column=1, columnspan=columns - 1, sticky="ew")
        if self.directory:
            self.choose = tkinter.ttk.Button(
                master,
                text="Choose...",
                command=lambda: self.value.set(tkinter.filedialog.askdirectory()),
            )
        else:
            self.choose = tkinter.ttk.Button(
                master,
                text="Choose...",
                command=lambda: self.value.set(tkinter.filedialog.askopenfilename()),
            )
        self.choose.grid(row=row + 1, column=1, sticky="w")
        if self.optional:
            self.clear = tkinter.ttk.Button(
                master, text="Clear", command=lambda: self.value.set("")
            )
            self.clear.grid(row=row + 1, column=2, sticky="e")
        return 2

    def validate(self):
        if not self.enabled:
            return True
        elif len(self.value.get()) == 0:
            if not self.optional:
                tkinter.messagebox.showwarning("", "{} not specified.".format(self.text))
                return False
            else:
                return True
        else:
            if os.path.exists(self.value.get()):
                if self.extensions is not None:
                    if any(
                        self.value.get().lower().endswith(x) for x in self.extensions
                    ):
                        return True
                    else:
                        tkinter.messagebox.showwarning(
                            "", "Invalid file extension " "for {}.".format(self.text)
                        )
                        return False
                else:  # no extension restriction
                    return True
            else:
                tkinter.messagebox.showwarning(
                    "", "{} file does not exist." "".format(self.text)
                )
                return False

    def enable(self):
        self.enabled = True
        self.entry.state(["!disabled"])
        self.choose.state(["!disabled"])
        if self.optional:
            self.clear.state(["!disabled"])

    def disable(self):
        self.enabled = False
        self.entry.state(["disabled"])
        self.choose.state(["disabled"])
        if self.optional:
            self.clear.state(["disabled"])


class StringEntry(MyEntry):
    """
    Creates a labeled Entry field for a string.

    *text* is the Label/error box text.
    """

    def __init__(self, text, cfg, key, optional=False):
        MyEntry.__init__(self, text, cfg, key, optional)

    def body(self, master, row, columns=DEFAULT_COLUMNS, **kwargs):
        """
        Place the required elements using the grid layout method.

        Returns the number of rows taken by this element.
        """
        label = tkinter.ttk.Label(master, text=self.text)
        label.grid(row=row, column=0, columnspan=1, sticky="e")
        self.entry = tkinter.ttk.Entry(master, textvariable=self.value)
        self.entry.grid(row=row, column=1, columnspan=columns - 1, sticky="ew")
        return 1


class IntegerEntry(MyEntry):
    """
    Creates a labeled Entry field for an integer.

    *text* is the Label/error box text.
    """

    def __init__(self, text, cfg, key, optional=False, minvalue=0):
        MyEntry.__init__(self, text, cfg, key, optional)
        self.minvalue = minvalue

    def body(self, master, row, columns=DEFAULT_COLUMNS, width=4, left=False, **kwargs):
        """
        Add the labeled entry to the Frame *master* using grid at *row*.

        *width* controls the width of the Entry.
        *left* is ``True`` if the Entry is to the left of the Label.
        *columns* is the number of columns in *master*.

        Returns the number of rows taken by this element.
        """
        if left:
            entry_column = 0
            entry_sticky = "e"
            entry_width = 1
            label_column = 1
            label_sticky = "w"
            label_width = columns - 1
        else:
            entry_column = 1
            entry_sticky = "w"
            entry_width = columns - 1
            label_column = 0
            label_sticky = "e"
            label_width = 1

        label = tkinter.ttk.Label(master, text=self.text)
        label.grid(
            row=row, column=label_column, columnspan=label_width, sticky=label_sticky
        )
        self.entry = tkinter.ttk.Entry(master, textvariable=self.value, width=width)
        self.entry.grid(
            row=row, column=entry_column, columnspan=entry_width, sticky=entry_sticky
        )
        return 1

    def validate(self):
        """
        Returns ``True`` if the value entered validates; else ``False``.

        If *self.optional* is ``True``, the field can be empty.
        Checks the *self.minvalue* that was passed on creation.
        """
        if not self.enabled:
            return True
        else:
            try:
                intvalue = int(self.value.get())
            except ValueError:
                if len(self.value.get()) == 0:
                    if not self.optional:
                        tkinter.messagebox.showwarning(
                            "", "{} not specified." "".format(self.text)
                        )
                        return False
                    else:
                        return True
                else:
                    tkinter.messagebox.showwarning(
                        "", "{} is not an integer." "".format(self.text)
                    )
                    return False
            else:
                if intvalue < self.minvalue:
                    tkinter.messagebox.showwarning(
                        "",
                        "{} lower than minimum value "
                        "({}).".format(self.text, self.minvalue),
                    )
                    return False
                else:
                    return True

    def apply(self):
        if self.enabled and len(self.value.get()) > 0:
            self.cfg[self.key] = int(self.value.get())
        else:
            self.cfg[self.key] = None



import logging
from .seqlib import SeqLib
from .barcode import BarcodeSeqLib
from .barcodemap import BarcodeMap
import pandas as pd
from .plots import barcodemap_plot
from matplotlib.backends.backend_pdf import PdfPages
import os.path


class BcidSeqLib(BarcodeSeqLib):
    """
    Class for counting data from barcoded sequencing libraries with non-variant
    identifiers. 
    Creating a :py:class:`BcidSeqLib` requires a valid *config* 
    object with an ``'barcodes'`` entry and information.

    The ``barcode_map`` keyword argument can be used to pass an existing 
    :py:class:`~seqlib.barcodemap.BarcodeMap`. Ensuring this is the 
    right :py:class:`~seqlib.barcodemap.BarcodeMap` is the responsibility 
    of the caller.
    """

    treeview_class_name = "Barcoded ID SeqLib"

    def __init__(self):
        BarcodeSeqLib.__init__(self)
        self.barcode_map = None
        self.identifier_min_count = 0
        self.add_label("identifiers")
        self.logger = logging.getLogger("{}.{}".format(__name__, self.__class__))

    def configure(self, cfg, barcode_map=None):
        """
        Set up the object using the config object *cfg*, usually derived from 
        a ``.json`` file.
        """
        BarcodeSeqLib.configure(self, cfg)
        self.logger = logging.getLogger(
            "{}.{} - {}".format(__name__, self.__class__.__name__, self.name)
        )
        try:
            if "min count" in cfg["identifiers"]:
                self.identifier_min_count = int(cfg["identifiers"]["min count"])

            if barcode_map is not None:
                if barcode_map.filename == cfg["barcodes"]["map file"]:
                    self.barcode_map = barcode_map
                else:
                    raise ValueError(
                        "Attempted to assign non-matching barcode map [{}]".format(
                            self.name
                        )
                    )
            else:
                self.barcode_map = BarcodeMap(
                    cfg["barcodes"]["map file"], is_variant=False
                )
        except KeyError as key:
            raise KeyError(
                "Missing required config value {key} [{name}]".format(
                    key=key, name=self.name
                )
            )

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for dumping to a config file.
        """
        cfg = BarcodeSeqLib.serialize(self)

        cfg["identifiers"] = dict()
        if self.identifier_min_count > 0:
            cfg["identifiers"]["min count"] = self.identifier_min_count

        if self.barcode_map is not None:  # required for creating new objects in GUI
            cfg["barcodes"]["map file"] = self.barcode_map.filename

        return cfg

    def calculate(self):
        """
        Counts the barcodes using :py:meth:`BarcodeSeqLib.count` and combines them into 
        identifier counts using the :py:class:`BarcodeMap`.
        """
        if not self.check_store("/main/identifiers/counts"):
            BarcodeSeqLib.calculate(self)  # count the barcodes
            df_dict = dict()
            barcode_identifiers = dict()

            self.logger.info("Converting barcodes to identifiers")
            # store mapped barcodes
            self.save_filtered_counts(
                "barcodes",
                "index in self.barcode_map.keys() & count >= self.barcode_min_count",
            )

            # count identifiers associated with the barcodes
            for bc, count in self.store["/main/barcodes/counts"].iterrows():
                count = count["count"]
                identifier = self.barcode_map[bc]
                try:
                    df_dict[identifier] += count
                except KeyError:
                    df_dict[identifier] = count
                barcode_identifiers[bc] = identifier

            # save counts, filtering based on the min count
            self.save_counts(
                "identifiers",
                {
                    k: v
                    for k, v in df_dict.items()
                    if v >= self.identifier_min_count
                },
                raw=False,
            )
            del df_dict

            # write the active subset of the BarcodeMap to the store
            barcodes = list(barcode_identifiers.keys())
            barcode_identifiers = pd.DataFrame(
                {"value": [barcode_identifiers[bc] for bc in barcodes]}, index=barcodes
            )
            del barcodes
            barcode_identifiers.sort_values("value", inplace=True)
            self.store.put(
                "/raw/barcodemap",
                barcode_identifiers,
                data_columns=barcode_identifiers.columns,
                format="table",
            )
            del barcode_identifiers

            # self.report_filter_stats()
            self.save_filter_stats()

    def make_plots(self):
        """
        Make plots for :py:class:`~seqlib.seqlib.BcidSeqLib` objects.

        Creates plot of the number of barcodes mapping to each identifier.
        """
        if self.plots_requested:
            SeqLib.make_plots(self)
            # open the PDF file
            pdf = PdfPages(os.path.join(self.plot_dir, "barcodes_per_identifier.pdf"))
            barcodemap_plot(self, pdf)
            pdf.close()




import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Circle
from matplotlib.lines import Line2D
from matplotlib.gridspec import GridSpec
from .plots import plot_cmaps, plot_colors

logger = logging.getLogger(__name__)

#: List of amino acids in row order for sequence-function maps.
AA_LIST = [
    "H",
    "K",
    "R",  # (+)
    "D",
    "E",  # (-)
    "C",
    "M",
    "N",
    "Q",
    "S",
    "T",  # Polar-neutral
    "A",
    "I",
    "L",
    "V",  # Non-polar
    "F",
    "W",
    "Y",  # Aromatic
    "G",
    "P",  # Unique
    "*",
]

#: List of tuples for amino acid physiochemical property groups.
#: Each tuple contains the label string and the corresponding start and end
#: indices in :py:const:`aa_list` (inclusive).
AA_LABEL_GROUPS = [
    ("(+)", 0, 2),
    ("(-)", 3, 4),
    ("Polar-neutral", 5, 10),
    ("Non-polar", 11, 14),
    ("Aromatic", 15, 17),
    ("Unique", 18, 19),
]

#: List of nucleotides in row order for sequence-function maps.
NT_LIST = ["A", "C", "G", "T"]


def parse_aa_list(fname):
    group_labels = list()
    groups = list()
    found_aa = list()
    with open(fname) as handle:
        for line in handle:
            if line.startswith("#"):
                continue
            if len(line) == 0:
                continue
            try:
                label, aa = line.split("\t")
            except ValueError:
                raise ValueError(
                    "Unexpected AA list file line format [{}]" "".format("SFMAP.PY")
                )
            aa = aa.split(",")
            aa = [x.strip() for x in aa]
            if any(x not in AA_LIST for x in aa):
                raise ValueError(
                    "Invalid amino acid in AA list file [{}]" "".format("SFMAP.PY")
                )
            else:
                group_labels.append(label)
                groups.append(aa)
                found_aa.extend(aa)
    if any(x not in found_aa for x in AA_LIST):
        raise ValueError(
            "Not all amino acids assigned in AA list file [{}]" "".format("SFMAP.PY")
        )
    if len(AA_LIST) != len(found_aa):
        raise ValueError(
            "Duplicate assignments in AA list file [{}]" "".format("SFMAP.PY")
        )
    pos = -1
    aa_label_groups = list()
    for i, label in enumerate(group_labels):
        aa_label_groups.append((label, pos + 1, pos + len(groups[i])))
        pos = pos + len(groups[i])

    return found_aa, aa_label_groups


def recentered_cmap(cmap, vmin, vmax):
    """
    Rescale the diverging color map *cmap* such that the center color is at 0
    in the data. Returns the rescaled cmap.

    Based on http://stackoverflow.com/a/20528097

    *cmap* is the color map to re-center. Should be a diverging brewer
    palette.

    *vmin* is the minimum value in the data being plotted with the *cmap*
    (must be negative).

    *vmax* is the maximum value in the data being plotted with the *cmap*
    (must be positive).

    """
    # regular index to compute the colors
    reg_index = np.linspace(0.0, 1.0, 257)

    # shifted index to match the data
    centerpoint = 1 - vmax / (vmax + abs(vmin))
    shift_index = np.hstack(
        [
            np.linspace(0.0, centerpoint, 128, endpoint=False),
            np.linspace(centerpoint, 1.0, 129, endpoint=True),
        ]
    )

    # re-map the colors at each index value using a dictionary
    cdict = {"red": [], "green": [], "blue": [], "alpha": []}
    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict["red"].append((si, r, r))
        cdict["green"].append((si, g, g))
        cdict["blue"].append((si, b, b))
        cdict["alpha"].append((si, a, a))

    # save the dictionary as a color map
    newcmap = LinearSegmentedColormap("recentered", cdict)

    return newcmap


def sfmap_axes(
    df,
    ax,
    style,
    wt,
    df_se=None,
    tall=False,
    colors=None,
    missing_color=None,
    vmin=None,
    vmax=None,
    vmax_se=None,
    show_positions=True,
    show_wt=True,
    show_variants=True,
    aa_list=AA_LIST,
    aa_label_groups=AA_LABEL_GROUPS,
):
    """
    Create heatmap of scores or counts for each position in *df* using the axes
    instance *ax*. Returns the |mpl_pcolormesh| (requred to create the color
    bar).

    Amino acids are ordered by their physiochemical properties:

    .. csv-table::
        :header: "Charge" , "Amino Acid" , "Abbreviation" , "Letter"

        Charged (+) , Histidine , His , H
        Charged (+) , Lysine , Lys , K
        Charged (+) , Arginine , Arg , R
        Charged (-) , Aspartate , Asp , D
        Charged (-) , Glutamate , Glu , E
        Polar-neutral , Cysteine , Cys , C
        Polar-neutral , Methionine , Met , M
        Polar-neutral , Asparagine , Asn , N
        Polar-neutral , Glutamine , Gln , Q
        Polar-neutral , Serine , Ser , S
        Polar-neutral , Threonine , Thr , T
        Non-polar , Alanine , Ala , A
        Non-polar , Glycine , Gly , G
        Non-polar , Isoleucine , Ile , I
        Non-polar , Leucine , Leu , L
        Non-polar , Proline , Pro , P
        Non-polar , Valine , Val , V
        Aromatic , Phenylalanine , Phe , F
        Aromatic , Tryptophan , Trp , W
        Aromatic , Tyrosine , Tyr , Y

    *df* is a DataFrame containing scores or counts. 
        Rows are labeled by the integer positions in the wild type sequence,
        numbered in ascending order. Columns are labeled by nucleotide or
        amino acid change.

    *ax* is the axes object used for the heatmap.

    *style* is one of ``'counts'``, ``'logcounts'``, or ``'scores'``.
        ``'counts'`` is used for plotting raw count data for library diversity
        maps.

        ``'logcounts'`` is also used for library diversity maps, but the counts
        are log10 transformed before plotting.

        ``'scores'`` is used for sequence-function maps.

    *wt* is the wild type sequence for the region included in *df*.

    *tall* sets the orientation of the heatmap. 
        ``True`` when the rows of the heatmap correspond to positions, which
        typically results in a plot that is taller than it is wide. Otherwise,
        the columns of the heatmap correspond to positions.

        The actual size of the heatmap is determined by the function that
        handles the returned axes object.

    *colors* is the name of the `matplotlib cmap`_ used.
        If *colors* is ``None``, the ``'counts'`` and ``'logcounts'`` *style*
        use the default sequential cmap and the ``'scores'`` *style* uses the
        default diverging cmap.

        When using the ``'scores'`` style, the cmap is automatically
        re-centered such that the center of the cmap corresponds to 0 in the
        data.

    *missing_color* is the color to use for missing data (``NaN``) in the plot.
        If *missing_color* is ``None``, the default missing color is used.

    *vmin* and *vmax* override the maximum and minimum values in *df*.
        These are used for plotting multiple plots with different ranges using
        the same color bar.

    *show_positions* sets the drawing of the integer position axis tick labels.

    *show_wt* sets the drawing of the wild type sequence axis tick labels.

    *show_variants* sets the drawing of the amino acid or nucleotide change
    axis tick labels.



    *df_se* contains the standard error of each item in *df*

    Standard errors are plotted as diagonal bars on each cell, scaled as a
    percentage of the highest standard error in the plot (or *vmax_se* if
    specified). To scale standard error bars to the highest standard error
    the dataset, calculate the maximum and send it to *vmax_se*. Standard
    errors that are less than 2% of the maximum are not plotted.
    """
    if style not in ("counts", "logcounts", "scores"):
        logger.warning("Invalid style specified for sfmap_axes")
        return None

    # rotate if necessary
    if not tall:
        df = df.transpose()
        if df_se is not None:
            df_se = df_se.transpose()

    # re-order the rows for plotting
    df = df.reindex(index=df.index[::-1])
    if df_se is not None:
        df_se = df_se.reindex(index=df_se.index[::-1])

    # log transform counts if necessary
    if style == "logcounts":
        df = np.log10(df)

    # mask NaN so we can color them separately
    masked = np.ma.array(df, mask=np.isnan(df))
    if df_se is not None:
        masked_se = np.ma.array(df_se, mask=np.isnan(df_se))

    # set colors and min/max data values
    if style == "counts" or style == "logcounts":
        try:
            if colors is not None:
                cmap = plt.get_cmap(colors)
            else:
                cmap = plt.get_cmap(plot_cmaps["sequential"])
        except ValueError:
            logger.warning(
                "Invalid sequential color map choice. " "Falling back to 'BuPu'"
            )
            cmap = plt.get_cmap("BuPu")
        if vmin is None:
            vmin = 0.0
        if vmax is None:
            vmax = masked.max()
    else:  # style == "scores"
        try:
            if colors is not None:
                cmap = plt.get_cmap(colors)
            else:
                cmap = plt.get_cmap(plot_cmaps["diverging"])
        except ValueError:
            logger.warning(
                "Invalid diverging color map choice. " "Falling back to 'RdYlBu_r'"
            )
            cmap = plt.get_cmap("RdYlBu_r")
        if vmin is None:
            vmin = masked.min()
        if vmax is None:
            vmax = masked.max()
        if vmin < 0.0 and vmax > 0.0:
            cmap = recentered_cmap(cmap, vmin, vmax)
        else:
            logger.warning("Unexpected range. Not recentering color map.")
    if missing_color is not None:
        cmap.set_bad(missing_color, 1.0)
    else:
        cmap.set_bad(plot_colors["missing"], 1.0)

    # draw the heatmap and get rid of extra whitespace
    mesh = ax.pcolormesh(masked, cmap=cmap, vmin=vmin, vmax=vmax)
    ax.set_xlim(0, len(df.columns))
    ax.set_ylim(0, len(df.index))

    # add marks on wild type positions
    wt = list(wt)
    if tall:
        wt_xy = list(zip((list(df.columns).index(x) for x in wt), reversed(list(range(len(wt))))))
    else:
        wt_xy = list(zip(list(range(len(wt))), (list(df.index).index(x) for x in wt)))
    for x, y in wt_xy:
        ax.add_patch(
            Circle(
                (x + 0.5, y + 0.5),
                0.1666,
                fill=True,
                facecolor="black",
                edgecolor="none",
                alpha=0.5,
            )
        )

    # add diagonal SE marks
    if df_se is not None:
        if vmax_se is None:
            vmax_se = masked_se.max()
        # rescale the SE's onto 0 .. 0.98
        # rescaling onto 0 .. 1.0 causes the corners to look funny
        masked_se = masked_se / vmax_se * 0.98
        for x in range(len(df.index)):
            for y in range(len(df.columns)):
                value = masked_se[x, y]
                if value and value >= 0.02:  # not masked, above threshold
                    corner_dist = (1.0 - value) / 2.0
                    diag = Line2D(
                        [y + corner_dist, y + 1 - corner_dist],
                        [x + corner_dist, x + 1 - corner_dist],
                        transform=ax.transData,
                        color="grey",
                    )
                    ax.add_line(diag)

    # calculate tick label positions
    ypos = np.arange(len(df.index)) + 0.5
    xpos = np.arange(len(df.columns)) + 0.5
    ax.set_yticks(ypos)
    ax.set_xticks(xpos)

    # create the secondary axis for the wild type sequence
    if tall:
        wt_ax = ax.twinx()
        wt_ax.set_yticks(ypos)
    else:
        wt_ax = ax.twiny()
        wt_ax.set_xticks(xpos)

    # add the integer position axis tick labels
    if show_positions:
        # add position axis label
        if tall:
            ax.set_ylabel("Position")
        else:
            ax.set_xlabel("Position")
        if tall:
            ax.set_yticklabels(list(df.index))
        else:
            ax.set_xticklabels(list(df.columns), rotation=90)

    # add the variant information
    # note this is added as text within the plot not tick labels
    if show_variants:
        if tall:
            # turn off existing xticklabels
            plt.setp(ax.get_xticklabels(), visible=False)
            # add fake xlabels as plot text
            for i, x in enumerate(list(df.columns)):
                ax.text(
                    i + 0.5,
                    len(df.index) + 0.4,
                    x,
                    horizontalalignment="center",
                    verticalalignment="center",
                    transform=ax.transData,
                )
            # add additional amino acid property information
            if list(df.columns) == aa_list:
                ax.set_ylim(0, len(df.index) + 2.0)
                wt_ax.set_ylim(0, len(df.index) + 2.0)
                for label, start, end in aa_label_groups:
                    if len(label) == 0:
                        continue
                    ax.text(
                        (end - start + 1) / 2.0 + start,
                        len(df.index) + 1.4,
                        label,
                        horizontalalignment="center",
                        verticalalignment="center",
                        transform=ax.transData,
                    )
                    bar = Line2D(
                        [start + 0.125, end + 1 - 0.125],
                        [len(df.index) + 0.9, len(df.index) + 0.9],
                        transform=ax.transData,
                        color="grey",
                    )
                    bar.set_clip_on(False)
                    ax.add_line(bar)
            # no amino acid properties, just add padding
            else:
                ax.set_ylim(0, len(df.index) + 1.0)
                wt_ax.set_ylim(0, len(df.index) + 1.0)
        else:
            # turn off existing yticklabels
            plt.setp(ax.get_yticklabels(), visible=False)
            # add fake ylabels as plot text
            for i, x in enumerate(list(df.index)):
                ax.text(
                    -0.4,
                    i + 0.5,
                    x,
                    horizontalalignment="center",
                    verticalalignment="center",
                    transform=ax.transData,
                )
            # add additional amino acid property information
            if list(df.index)[::-1] == aa_list:
                ax.set_xlim(-2, len(df.columns))
                wt_ax.set_xlim(-2, len(df.columns))
                for label, start, end in aa_label_groups:
                    if len(label) == 0:
                        continue
                    ax.text(
                        -1,
                        len(df.index) - ((end - start + 1) / 2.0 + start),
                        label,
                        rotation="vertical",
                        verticalalignment="center",
                        horizontalalignment="right",
                        transform=ax.transData,
                    )
                    bar = Line2D(
                        [-0.9, -0.9],
                        [
                            len(df.index) - (start + 0.125),
                            len(df.index) - (end + 1 - 0.125),
                        ],
                        transform=ax.transData,
                        color="grey",
                    )
                    bar.set_clip_on(False)
                    ax.add_line(bar)
            # no amino acid properties, just add padding
            else:
                ax.set_xlim(-1, len(df.columns))
                wt_ax.set_xlim(-1, len(df.columns))

    # add the wild type sequence labels
    if show_wt:
        if tall:
            wt_ax.set_yticklabels(wt[::-1])
            wt_ax.set_ylabel("Wild Type Sequence")
        else:
            wt_ax.set_xticklabels(wt)
            wt_ax.set_xlabel("Wild Type Sequence")

    # hide the tick marks
    # needs to be done after the ax.twiny() or ax.twinx() call for some reason
    ax.tick_params(bottom=False, top=False, left=False, right=False)
    ax.set_frame_on(False)
    wt_ax.tick_params(bottom=False, top=False, left=False, right=False)
    wt_ax.set_frame_on(False)

    return mesh


def sfmap_plot(
    df, pdf, style, wt, dimensions, df_se=None, title=None, show_colorbar=True, **kwargs
):
    """
    Create heatmap of scores or counts for each position in *df* and save it to
    the file *pdf*.

    *df* is a DataFrame containing scores or counts.
        Rows are labeled by the integer positions in the wild type sequence,
        numbered in ascending order. Columns are labeled by nucleotide or
        amino acid change.

    *pdf* is the open PdfPages file object.

    *style* is one of ``'counts'``, ``'logcounts'``, or ``'scores'``.
        ``'counts'`` is used for plotting raw count data for library diversity
        maps.

        ``'logcounts'`` is also used for library diversity maps, but the counts
        are log10 transformed before plotting.

        ``'scores'`` is used for sequence-function maps.

    *wt* is the wild type sequence for the region included in *df*.

    *dimensions* is either ``'tall'`` or ``'wide'``, or a two-tuple of floats
    giving the figure size in inches.

        ``'tall'`` will create a figure 8.5 inches wide that is tall enough to
        make all cells in the heatmap roughly square.

        ``'wide'`` will create a figure 8.5 inches tall that is wide enough to
        make all cells in the heatmap roughly square.

        The two-tuple of floats specifies the exact figure size (including
        title and color bar, if requested) in inches. The long dimension is
        determined based on the ratio between width and height.

        .. note:: Specifying a figure size may cause the axis labels to not \
        display properly.

    *title* is the figure title. It can include ``'\\n'`` characters for a
    multi-line title.

    *show_colorbar* is ``True`` to display a color bar, or ``False`` to not
    display a color bar. If the figure is in tall mode, the color bar appears
    along the bottom of the plot, otherwise it appears in the right side.

    *kwargs* is the dictionary of additional optional or required parameters to
    :py:func:`sfmap_axes`.

    """
    if style not in ("counts", "logcounts", "scores"):
        logger.warning("Invalid style specified for sfmap_axes.")
        return None

    # dimensions set explicitly
    if len(dimensions) == 2:
        try:
            dimx, dimy = [float(n) for n in dimensions]
        except ValueError:
            logger.warning("Invalid sequence-function map dimensions.")
            return None
    else:
        # scaling based on mode
        if dimensions == "tall":
            # headings + pos_ax + wt_ax
            colx = len(df.columns)
            # rows + cbar_ax + aa_headings + title_ax
            coly = len(df.index) + 2 + 2 + 4
        elif dimensions == "wide":
            # rows + aa_headings + cbar_ax
            colx = len(df.index) + 2 + 2
            # headings + pos_ax + wt_ax + title_ax
            coly = len(df.columns) + 2
        else:
            logger.warning("Invalid sequence-function map scaling mode.")
            return None
        base_scale = 8.5 / 21
        dimx = colx * base_scale
        dimy = coly * base_scale

    tall = dimy > dimx

    # create the figure
    fig = plt.figure()
    fig.set_size_inches(dimx, dimy)

    # set up the grid for the figure
    if tall:
        if title is not None and show_colorbar:
            grid = GridSpec(
                3,
                1,
                height_ratios=[1, len(df.index) + 2, 1],
                hspace=1.0 / len(df.index),
            )
            title_ax = plt.subplot(grid[0])
            mesh_ax = plt.subplot(grid[1])
            cbar_ax = plt.subplot(grid[2])
        elif title is not None:  # no colorbar
            grid = GridSpec(2, 1, height_ratios=[1, len(df.index)])
            title_ax = plt.subplot(grid[0])
            mesh_ax = plt.subplot(grid[1])
            cbar_ax = None
        elif show_colorbar:  # no title
            grid = GridSpec(2, 1, height_ratios=[len(df.index), 1])
            title_ax = None
            mesh_ax = plt.subplot(grid[0])
            cbar_ax = plt.subplot(grid[1])
        else:  # only the heatmap
            grid = GridSpec(1, 1)
            title_ax = None
            mesh_ax = plt.subplot(grid[0])
            cbar_ax = None
    else:
        if title is not None and show_colorbar:
            grid = GridSpec(
                2,
                2,
                height_ratios=[1, len(df.columns)],
                width_ratios=[len(df.index), 1],
                wspace=1.0 / len(df.index),
            )
            title_ax = plt.subplot(grid[0, :])
            mesh_ax = plt.subplot(grid[1, 0])
            cbar_ax = plt.subplot(grid[1, 1])
        elif title is not None:  # no colorbar
            grid = GridSpec(2, 1, height_ratios=[1, len(df.index)])
            title_ax = plt.subplot(grid[0])
            mesh_ax = plt.subplot(grid[1])
            cbar_ax = None
        elif show_colorbar:  # no title
            grid = GridSpec(1, 2, width_ratios=[len(df.columns), 1])
            title_ax = None
            mesh_ax = plt.subplot(grid[0])
            cbar_ax = plt.subplot(grid[1])
        else:  # only the heatmap
            grid = GridSpec(1, 1)
            title_ax = None
            mesh_ax = plt.subplot(grid[0])
            cbar_ax = None

    # create the heatmap
    mesh = sfmap_axes(df, mesh_ax, style, wt, df_se=df_se, tall=tall, **kwargs)

    # add the colorbar axis
    if cbar_ax is not None:
        if tall:
            cbar = fig.colorbar(mesh, cax=cbar_ax, orientation="horizontal")
        else:
            cbar = fig.colorbar(mesh, cax=cbar_ax)
        if style == "logcounts":
            cbar.set_label("log10(Count)")
        elif style == "counts":
            cbar.set_label("Count")
        else:  # style == "scores"
            cbar.set_label("Score")
        cbar.outline.set_visible(False)
        cbar.ax.tick_params(bottom=False, top=False, left=False, right=False)

    # add the title
    if title_ax is not None:
        title_ax.text(s=title, x=0.5, y=1.0, ha="center", va="top", size="large")
        title_ax.axis("off")

    # save the heatmap
    pdf.savefig(fig, bbox_inches="tight", pad_inches=0.05)
    plt.close(fig)



import pandas as pd
import numpy as np
import logging
from .constants import WILD_TYPE_VARIANT
import collections
from .variant import mutation_count, re_protein, re_coding, re_noncoding
from .barcodemap import re_barcode, re_identifier
from .constants import AA_CODES
from .storemanager import ELEMENT_LABELS
from .sfmap import AA_LIST, NT_LIST

logger = logging.getLogger(__name__)

SingleMut = collections.namedtuple("SingleMut", ["pre", "post", "pos", "key"])


def validate_index(index, element):
    """
    Return a boolean list for which index values are valid for the given
    element type.
    """
    if element not in ELEMENT_LABELS:
        raise ValueError("Invalid element label '{}'".format(element))

    if element == "barcodes":
        retval = [re_barcode.match(x) is not None for x in index]
    elif element == "identifiers":
        retval = [re_identifier.match(x) is not None for x in index]
    elif element == "variants":
        pass
    elif element == "synonymous":
        pass
    else:
        raise NotImplementedError("Unimplemented element type '{}'" "".format(element))


def single_mutation_index(index):
    """
    Return a filtered pandas Index containing only single mutations. Filtering
    also removes unrecognized amino acids (denoted by ``"???"``) caused by
    some indels.

    *index* the index to be filtered for single mutations.
    """
    return pd.Index(x for x in index if mutation_count(x) == 1)


def filter_coding_index(index):
    """
    Return a filtered pandas Index with any unrecognized amino acids (denoted
    by ``"???"``) removed. These are caused by some frame shift mutations.

    *index* the index to be filtered.
    """
    return pd.Index(x for x in index if "???" not in x)


def single_mutations_to_tuples(index):
    """
    Return a list of SingleMut namedtuples for each single mutation in the
    *index*. The type of index (noncoding DNA, coding DNA, or protein) is
    automatically detected.

    Position value in the tuple is stored as an integer.

    If the *index* is a protein index, the amino acids are referred to by
    single-letter codes not three-letter codes.

    *index* is the index to convert to SingleMut tuples.

    Raises a ValueError if non-single mutations are included in *index*.

    Raises a ValueError if one of the *index* entries cannot be parsed.

    Raises an IndexError if the *index* is empty.
    """
    if any(mutation_count(x) != 1 for x in index):
        raise ValueError(
            "Non-single mutations cannot be converted into " "SingleMut tuples."
        )

    # identify the type of index
    try:
        if re_noncoding.match(index[0]):
            is_protein = False
            expression = re_noncoding
        elif re_coding.match(index[0]):
            is_protein = False
            expression = re_coding
        elif re_protein.match(index[0]):
            is_protein = True
            expression = re_protein
        else:
            raise ValueError("Unrecognized HGVS string.")
    except IndexError:
        raise IndexError("Cannot convert empty index to tuples.")

    # perform the regular expression matches and create the SingleMut tuples
    tuples = list()
    for x in index:
        m = expression.match(x)
        if m is None:
            raise ValueError("Unrecognized HGVS string.")
        else:
            if is_protein:  # convert to single-letter amino acid code
                tuples.append(
                    SingleMut(
                        AA_CODES[m.group("pre")],
                        AA_CODES[m.group("post")],
                        int(m.group("pos")),
                        m.group("match"),
                    )
                )
            else:
                tuples.append(
                    SingleMut(
                        m.group("pre"),
                        m.group("post"),
                        int(m.group("pos")),
                        m.group("match"),
                    )
                )

    return tuples


def fill_position_gaps(positions, gap_size):
    """
    Create a list of integer positions with gaps filled in. Used by
    :py:func:`singleton_dataframe`.

    Args:
        positions (list): integer positions
        gap_size (int): maximum length of gap that will be filled

    Returns:
        list: sorted list of unique integer positions with gaps filled
    """
    if len(positions) == 0:
        raise ValueError("Empty positions list.")

    # uniqify and sort
    positions = sorted(list(set(positions)))

    # fill in short gaps
    fill = set()
    for i in range(len(positions) - 1):
        delta = positions[i + 1] - positions[i]
        if delta > 1 and delta <= gap_size:
            fill.update(positions[i] + n + 1 for n in range(delta))
    fill.update(positions)

    return sorted(list(fill))


def singleton_dataframe(
    values, wt, gap_size=5, coding=True, plot_wt_score=True, aa_list=AA_LIST
):
    """
    Prepare data for plotting as a sequence-function map. Returns a data frame
    suitable for plotting as heat map data and a wild type sequence extracted
    from the variant information.

    The type of variants stored is automatically detected, and the index will
    be filtered for single mutations.

    The data frame has amino acids or nucleotides as columns and positions with
    rows. If there are no mutations at a given position, it will not appear in
    the data frame unless this gap is filled with rows containing no data. The
    wild type sequence entry for these rows will be blank.

    Args:
        values (|pd_Series|): data values (typically scores or counts)

        wt (WildTypeSequence): wild type for the data

        gap_size (int): maximum length of missing data gap that will be filled

        coding (bool): True for amino acid data, False for nucleotide

        plot_wt_score (bool): True if the wild type positions should have the
            wild type score, False if they should be missing

    Returns:
        tuple: two-element tuple containing a |pd_DataFrame| filled with the
        data values and a list of single-character wild type values
    """
    if len(values.index) == 0:
        raise ValueError(
            "Cannot process an empty data frame [{}]".format(wt.parent_name)
        )

    # save the wild type score for later
    if plot_wt_score:
        try:
            wt_score = values[WILD_TYPE_VARIANT]
        except KeyError:
            logger.warning("Wild type score not measured, will be missing in " "plots")
            wt_score = np.nan

    # select only rows with singleton mutations
    values = values[filter_coding_index(single_mutation_index(values.index))]

    # parse out the information from the index
    index_tuples = single_mutations_to_tuples(values.index)

    # create and populate the DataFrame
    # get sorted, unique list of positions that have a mutation
    positions = fill_position_gaps([x.pos for x in index_tuples], gap_size=gap_size)
    # initialize the DataFrame
    if coding:
        columns = aa_list
    else:
        columns = NT_LIST
    frame = pd.DataFrame(np.nan, columns=columns, index=positions)
    # populate the DataFrame
    for x in index_tuples:
        frame.loc[x.pos, x.post] = values.loc[x.key]

    # create a dictionary of position->nucleotide/amino acid
    wt_dict = dict(wt.position_tuples(protein=coding))

    # convert subset of the wild type dictionary into sequence
    try:
        wt_sequence = "".join(wt_dict[x] for x in positions)
    except KeyError:
        raise ValueError("Inconsistent wild type positions [{}]".format(wt.parent_name))

    # double-check that the wild type is consistent with the data frame
    for x in index_tuples:
        if x.pos in wt_dict:
            if x.pre != wt_dict[x.pos]:
                raise ValueError(
                    "Inconsistent wild type sequence [{}]".format(wt.parent_name)
                )

    # add wild type scores if desired
    if plot_wt_score:
        for p in positions:
            frame.loc[p, wt_dict[p]] = wt_score

    return (frame, wt_sequence)




import tkinter as tk
import tkinter.ttk
import tkinter.filedialog
import tkinter.messagebox
import platform
import json
from ..config_check import is_seqlib, is_experiment, is_selection, seqlib_type
from .delete_dialog import DeleteDialog
from .seqlib_apply_dialog import SeqLibApplyDialog
from .edit_dialog import EditDialog, clear_nones
from .create_root_dialog import CreateRootDialog
from .create_seqlib_dialog import CreateSeqLibDialog
from .runner_window import RunnerWindow, RunnerSavePrompt
from ..experiment import Experiment
from ..condition import Condition
from ..selection import Selection
from ..barcode import BarcodeSeqLib
from ..barcodevariant import BcvSeqLib
from ..barcodeid import BcidSeqLib
from ..basic import BasicSeqLib
from ..idonly import IdOnlySeqLib
from ..overlap import OverlapSeqLib
from ..seqlib import SeqLib
from ..storemanager import SCORING_METHODS, LOGR_METHODS


#: map class names to class definitions to avoid use of globals()
SEQLIB_CLASSES = {
    "BarcodeSeqLib": BarcodeSeqLib,
    "BcvSeqLib": BcvSeqLib,
    "BcidSeqLib": BcidSeqLib,
    "BasicSeqLib": BasicSeqLib,
    "IdOnlySeqLib": IdOnlySeqLib,
    "OverlapSeqLib": OverlapSeqLib,
}


def write_json(d, handle):
    """
    Write the contents of dictionary *d* to an open file *handle* in json format.

    *d* is passed to ``clear_nones`` before output.
    """
    json.dump(clear_nones(d), handle, indent=2, sort_keys=True)


class Configurator(tk.Tk):
    def __init__(self, version):
        tk.Tk.__init__(self)
        self.title("Enrich 2 Configurator v{}".format(version))

        self.treeview = None
        self.treeview_popup_target_id = None

        self.cfg_file_name = tk.StringVar()

        self.element_dict = dict()
        self.root_element = None

        # analysis options
        self.scoring_method = tk.StringVar()
        self.logr_method = tk.StringVar()
        self.force_recalculate = tk.BooleanVar()
        self.component_outliers = tk.BooleanVar()
        self.plots_requested = tk.BooleanVar()
        self.tsv_requested = tk.BooleanVar()

        # allow resizing
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        # create UI elements
        self.create_main_frame()
        self.create_menubar()
        self.create_treeview_context_menu()

    def create_treeview_context_menu(self):
        self.treeview_popup = tk.Menu(self, tearoff=0)
        self.treeview_popup.add_command(
            label="Apply FASTQ...", command=self.apply_seqlib_fastq
        )

    def apply_seqlib_fastq(self):
        SeqLibApplyDialog(self, self, self.treeview_popup_target_id)

    def treeview_context_menu(self, click):
        target = self.treeview.identify_row(click.y)
        if target != "":
            self.treeview_popup_target_id = target
            self.treeview_popup.post(click.x_root, click.y_root)
        self.treeview_popup_target_id = None

    def create_main_frame(self):
        # Frame for the Treeview and New/Edit/Delete buttons
        main = tkinter.ttk.Frame(self, padding=(3, 3, 12, 12))
        main.rowconfigure(0, weight=1)
        main.columnconfigure(0, weight=1)
        main.grid(row=0, column=0, sticky="nsew")

        # Frame for the Treeview and its scrollbars
        tree_frame = tkinter.ttk.Frame(main, padding=(3, 3, 12, 12))
        tree_frame.rowconfigure(0, weight=1)
        tree_frame.rowconfigure(1, weight=0)
        tree_frame.columnconfigure(0, weight=1)
        tree_frame.columnconfigure(1, weight=0)
        tree_frame.grid(row=0, column=0, sticky="nsew")

        # Treeview with column headings
        self.treeview = tkinter.ttk.Treeview(tree_frame)
        self.treeview["columns"] = ("class", "barcodes", "variants")
        self.treeview.column("class", width=120)
        self.treeview.heading("class", text="Type")
        self.treeview.column("barcodes", width=25, stretch=tk.NO, anchor=tk.CENTER)
        self.treeview.heading("barcodes", text="BC")
        self.treeview.column("variants", width=25, stretch=tk.NO, anchor=tk.CENTER)
        self.treeview.heading("variants", text="V")
        self.treeview.grid(row=0, column=0, sticky="nsew")

        # Treeview context menu bindings
        self.treeview.bind("<Button-2>", self.treeview_context_menu)

        # Treeview scrollbars
        tree_ysb = tk.Scrollbar(
            tree_frame, orient="vertical", command=self.treeview.yview
        )
        tree_xsb = tk.Scrollbar(
            tree_frame, orient="horizontal", command=self.treeview.xview
        )
        tree_ysb.grid(row=0, column=1, sticky="nsw")
        tree_xsb.grid(row=1, column=0, sticky="ewn")
        self.treeview.config(yscroll=tree_ysb.set, xscroll=tree_xsb.set)

        # Frame for New/Edit/Delete buttons
        button_frame = tkinter.ttk.Frame(main, padding=(3, 3, 12, 12))
        button_frame.grid(row=1, column=0)
        new_button = tkinter.ttk.Button(
            button_frame, text="New...", command=self.new_button_press
        )
        new_button.grid(row=0, column=0)
        edit_button = tkinter.ttk.Button(
            button_frame, text="Edit...", command=self.edit_button_press
        )
        edit_button.grid(row=0, column=1)
        delete_button = tkinter.ttk.Button(
            button_frame, text="Delete", command=self.delete_button_press
        )
        delete_button.grid(row=0, column=2)

        # Frame for Analysis Options
        options_frame = tkinter.ttk.Frame(main, padding=(3, 3, 12, 12))
        options_frame.grid(row=0, column=1, rowspan=2, sticky="nsew")

        row = 0
        heading = tkinter.ttk.Label(options_frame, text="Analysis Options")
        heading.grid(column=0, row=row)
        row += 1

        scoring_heading = tkinter.ttk.Label(options_frame, text="Scoring Method")
        scoring_heading.grid(column=0, row=row)
        row += 1
        for i, k in enumerate(SCORING_METHODS.keys()):
            rb = tkinter.ttk.Radiobutton(
                options_frame,
                text=SCORING_METHODS[k].title(),
                variable=self.scoring_method,
                value=k,
            )
            rb.grid(column=0, row=row, sticky="w")
            row += 1
            if i == 0:
                rb.invoke()

        logr_heading = tkinter.ttk.Label(options_frame, text="Normalization Method")
        logr_heading.grid(column=0, row=row)
        row += 1
        for i, k in enumerate(LOGR_METHODS.keys()):
            rb = tkinter.ttk.Radiobutton(
                options_frame,
                text=LOGR_METHODS[k].title(),
                variable=self.logr_method,
                value=k,
            )
            rb.grid(column=0, row=row, sticky="w")
            row += 1
            if i == 0:
                rb.invoke()

        other_heading = tkinter.ttk.Label(options_frame, text="Other Options")
        other_heading.grid(column=0, row=row)
        row += 1

        # force recalculate
        force_recalculate = tkinter.ttk.Checkbutton(
            options_frame, text="Force Recalculation", variable=self.force_recalculate
        )
        force_recalculate.grid(column=0, row=row, sticky="w")
        row += 1

        # component outliers
        component_outliers = tkinter.ttk.Checkbutton(
            options_frame,
            text="Component Outlier Statistics",
            variable=self.component_outliers,
        )
        component_outliers.grid(column=0, row=row, sticky="w")
        row += 1

        # make plots
        plots_requested = tkinter.ttk.Checkbutton(
            options_frame, text="Make Plots", variable=self.plots_requested
        )
        plots_requested.grid(column=0, row=row, sticky="w")
        plots_requested.invoke()
        row += 1

        # write tsv
        tsv_requested = tkinter.ttk.Checkbutton(
            options_frame, text="Write TSV Files", variable=self.tsv_requested
        )
        tsv_requested.grid(column=0, row=row, sticky="w")
        tsv_requested.invoke()
        row += 1

        go_button = tkinter.ttk.Button(
            options_frame, text="Run Analysis", command=self.go_button_press
        )
        go_button.grid(column=0, row=row, sticky="sew")

    def go_button_press(self):
        if self.root_element is None:
            tkinter.messagebox.showwarning("", "No experimental design specified.")
        else:
            RunnerSavePrompt(self)
            RunnerWindow(self)

    def create_new_element(self):
        """
        Create and return a new element based on the current selection.

        This element is not added to the treeview.
        """
        element = None
        parent_element = self.get_focused_element()
        if isinstance(parent_element, Experiment):
            element = Condition()
            element.parent = parent_element
        elif isinstance(parent_element, Condition):
            element = Selection()
            element.parent = parent_element
        elif isinstance(parent_element, Selection):
            element = CreateSeqLibDialog(self).element_type()
            element.parent = parent_element
        elif isinstance(parent_element, SeqLib):
            # special case: creates a copy of the selected SeqLib as a sibling
            # don't allow creation of a second orphan SeqLib
            if parent_element.parent is None:
                tkinter.messagebox.showerror(None, "Cannot create multiple orphan SeqLibs.")
                return None
            element = type(parent_element)()
            element.configure(parent_element.serialize())
            element.parent = parent_element.parent
            # clear out the seqlib-specific values
            element.name = None
            element.timepoint = None
            element.counts_file = None
            if isinstance(element, OverlapSeqLib):
                element.forward = None
                element.reverse = None
            else:
                element.reads = None
        else:
            raise ValueError(
                "Unrecognized parent object type '{}'".format(type(parent_element))
            )
        return element

    def new_button_press(self):
        if self.treeview.focus() == "" and self.root_element is not None:
            tkinter.messagebox.showwarning(None, "No parent element selected.")
        else:
            if self.treeview.focus() == "" and self.root_element is None:
                element = CreateRootDialog(self).element
                if isinstance(element, SeqLib):
                    EditDialog(self, self, element)
                self.root_element = element
            else:
                element = self.create_new_element()
                if element is None:
                    return
                EditDialog(self, self, element)

            # refresh the treeview and re-assign treeview id's
            self.refresh_treeview()

            # select the newly added element if it was successfully added
            if element.treeview_id in list(self.element_dict.keys()):
                self.treeview.focus(element.treeview_id)
                self.treeview.selection_set(element.treeview_id)
            else:
                if element.parent is not None:
                    self.treeview.focus(element.parent.treeview_id)
                    self.treeview.selection_set(element.parent.treeview_id)
                del element

    def edit_button_press(self):
        if self.treeview.focus() == "":
            tkinter.messagebox.showwarning(None, "No element selected.")
        else:
            EditDialog(self, self, self.get_focused_element())

    def delete_button_press(self):
        if self.treeview.focus() == "":
            tkinter.messagebox.showwarning(None, "No element selected.")
        else:
            DeleteDialog(self, self)

    def create_menubar(self):
        # make platform-specific keybinds
        if platform.system() == "Darwin":
            accel_string = "Command+"
            accel_bind = "Command-"
        else:
            accel_string = "Ctrl+"
            accel_bind = "Control-"

        # create the menubar
        menubar = tk.Menu(self)

        # file menu
        filemenu = tk.Menu(menubar, tearoff=0)
        filemenu.add_command(
            label="Open...",
            accelerator="{}O".format(accel_string),
            command=self.menu_open,
        )
        filemenu.add_command(
            label="Save", accelerator="{}S".format(accel_string), command=self.menu_save
        )
        filemenu.add_command(
            label="Save As...",
            accelerator="{}Shift+S".format(accel_string),
            command=self.menu_saveas,
        )
        menubar.add_cascade(label="File", menu=filemenu)

        # edit menu
        filemenu = tk.Menu(menubar, tearoff=0)
        filemenu.add_command(
            label="Select All",
            accelerator="{}A".format(accel_string),
            command=self.menu_selectall,
        )
        menubar.add_cascade(label="Edit", menu=filemenu)

        # add the menubar
        self.config(menu=menubar)

        # add file menu keybinds
        self.bind("<{}o>".format(accel_bind), lambda event: self.menu_open())
        self.bind("<{}s>".format(accel_bind), lambda event: self.menu_save())
        self.bind("<{}Shift-s>".format(accel_bind), lambda event: self.menu_saveas())

        # add edit menu keybinds
        self.bind("<{}a>".format(accel_bind), lambda event: self.menu_selectall())

    def menu_open(self):
        fname = tkinter.filedialog.askopenfilename()
        if len(fname) > 0:  # file was selected
            try:
                with open(fname, "r") as handle:
                    cfg = json.load(handle)
            except ValueError:
                tkinter.messagebox.showerror(None, "Failed to parse config file.")
            except IOError:
                tkinter.messagebox.showerror(None, "Could not read config file.")
            else:
                if is_experiment(cfg):
                    obj = Experiment()
                elif is_selection(cfg):
                    obj = Selection()
                elif is_seqlib(cfg):
                    obj = SEQLIB_CLASSES[seqlib_type(cfg)]()
                else:
                    tkinter.messagebox.showerror(None, "Unrecognized config format.")
                    return
                obj.output_dir_override = False
                try:
                    obj.configure(cfg)
                except Exception as e:
                    tkinter.messagebox.showerror(
                        None, "Failed to process config file:\n{}".format(e)
                    )
                else:
                    self.root_element = obj
                    self.cfg_file_name.set(fname)
                    self.refresh_treeview()

    def menu_save(self):
        if len(self.cfg_file_name.get()) == 0:
            self.menu_saveas()
        elif self.root_element is None:
            tkinter.messagebox.showwarning(None, "Cannot save empty configuration.")
        else:
            try:
                with open(self.cfg_file_name.get(), "w") as handle:
                    write_json(self.root_element.serialize(), handle)
            except IOError:
                tkinter.messagebox.showerror(None, "Failed to save config file.")
            else:
                tkinter.messagebox.showinfo(
                    None, "Save successful:\n{}".format(self.cfg_file_name.get())
                )

    def menu_saveas(self):
        if self.root_element is None:
            tkinter.messagebox.showwarning(None, "Cannot save empty configuration.")
        else:
            fname = tkinter.filedialog.asksaveasfilename()
            if len(fname) > 0:  # file was selected
                try:
                    with open(fname, "w") as handle:
                        write_json(self.root_element.serialize(), handle)
                except IOError:
                    tkinter.messagebox.showerror(None, "Failed to save config file.")
                else:
                    self.cfg_file_name.set(fname)
                    tkinter.messagebox.showinfo(
                        None, "Save successful:\n{}".format(self.cfg_file_name.get())
                    )

    def menu_selectall(self):
        """
        Add all elements in the Treeview to the selection.
        """
        for k in list(self.element_dict.keys()):
            self.treeview.selection_add(k)

    def delete_element(self, tree_id):
        """
        Delete element with Treeview id *tree_id* from the tree, from the element
        dictionary, and from the associated data structure. Recursively
        deletes all children of *tree_id*.

        The tree should be refreshed using :py:meth:`refresh_tree` after
        each deletion. This is the responsibility of the caller.

        """
        # if self.treeview.exists(tree_id):
        if tree_id in self.element_dict:
            # recursively delete children
            if self.element_dict[tree_id].children is not None:
                for child in self.element_dict[tree_id].children:
                    self.delete_element(child.treeview_id)

            # check if deleting the root element
            if self.root_element.treeview_id == tree_id:
                # clear the root element
                self.root_element = None
            else:
                try:
                    # remove the element from its parent's list of children
                    self.element_dict[tree_id].parent.remove_child_id(tree_id)
                except AttributeError:
                    raise AttributeError("Non-root element lacks proper parent")

            # delete the element from the dictionary
            del self.element_dict[tree_id]

    def refresh_treeview(self):
        """
        Clears the Treeview and repopulates it with the current contents of the tree.
        """
        # clear the entries in the Treeview
        for x in self.treeview.get_children():
            self.treeview.delete(x)

        # clear the id-element dictionary
        # elements may be given new id's after repopulation
        self.element_dict.clear()

        # repopulate
        if self.root_element is not None:
            self.populate_tree(self.root_element)

    def set_treeview_properties(self, element):
        """
        Set the information text for the Treeview *element*.
        """
        # set class property
        self.treeview.set(element.treeview_id, "class", element.treeview_class_name)

        # add the check marks for barcodes/variants
        if "variants" in element.labels:
            self.treeview.set(element.treeview_id, "variants", "\u2713")
        else:
            self.treeview.set(element.treeview_id, "variants", "")
        if "barcodes" in element.labels:
            self.treeview.set(element.treeview_id, "barcodes", "\u2713")
        else:
            self.treeview.set(element.treeview_id, "barcodes", "")

        self.treeview.set(element.treeview_id, "class", element.treeview_class_name)

    def populate_tree(self, element, parent_id=""):
        """
        Recursively populate the Treeview.

        Also populates the *id_cfgstrings*.
        """
        # insert into the Treeview
        element.treeview_id = self.treeview.insert(
            parent_id, "end", text=element.name, open=True
        )
        # add id-element pair to dictionary
        self.element_dict[element.treeview_id] = element
        # set information fields
        self.set_treeview_properties(element)

        # populate for children
        if element.children is not None:
            for child in element.children:
                self.populate_tree(child, parent_id=element.treeview_id)

    def get_element(self, treeview_id):
        """
        Returns the element with *treeview_id*.
        """
        return self.element_dict[treeview_id]

    def get_focused_element(self):
        """
        Returns the element that is currently being focused in the Treeview.

        Returns ``None`` if nothing is focused.
        """
        if self.treeview.focus() != "":
            return self.get_element(self.treeview.focus())
        else:
            return None

    def get_selected_elements(self):
        """
        Returns a list of elements that are currently selected in the Treeview.

        If no elements are selected, it returns an empty list.
        """
        return [self.get_element(x) for x in self.treeview.selection()]



"""z-score helper functions.

This module contains functions for calculating z-scores and corresponding
p-values for pairs of scores.
"""

import numpy as np
import scipy.stats as stats


def ztest_pair(df1, df2):
    """z-test for elements in two data frames.

    Takes two data frames with ``'score'`` and ``'SE'`` columns and returns
    a new data frame containing the scores and standard errors and result for
    pairwise comparison of elements in both input data frames.

    Parameters
    ----------
    df1 : pandas.DataFrame
        The first data frame. Must contain ``'score'`` and ``'SE'`` columns.

    df2 : pandas.DataFrame
        The second data frame. Must contain ``'score'`` and ``'SE'`` columns.

    Returns
    -------
    pandas.DataFrame
        Result data frame containing scores (``'score1'`` and ``'score2'``),
        standard errors (``'SE1'``, ``'SE2'``), z-scores (``'z'``), and
        p-values (``'pvalue_raw'``) for each element found in both data frames.

    Raises
    ------
    To be added.
    """
    shared = df1.loc[:, ("score", "SE")].merge(
        df2.loc[:, ("score", "SE")],
        how="inner",
        left_index=True,
        right_index=True,
        suffixes=("1", "2"),
    )
    shared["z"] = np.abs(shared["score1"] - shared["score2"]) / np.sqrt(
        shared["SE1"] ** 2 + shared["SE2"] ** 2
    )
    shared["pvalue_raw"] = 2 * stats.norm.sf(shared["z"])
    return shared


def ztest_single(df, score, se):
    """z-test for comparing elements in a data frame to a single score and SE.

    Takes a data frames with ``'score'`` and ``'SE'`` columns and a score and
    standard error to compare them to and returns a new data frame containing
    the scores, standard errors, and result for the pairwise comparisons.

    Parameters
    ----------
    df : pandas.DataFrame
        The data frame. Must contain ``'score'`` and ``'SE'`` columns.

    score : float
        Score used for comparison to elements in the data frame.

    se : float
        Standard error used for comparison to elements in the data frame.

    Returns
    -------
    pandas.DataFrame
        Result data frame containing score (``'score'``), standard error
        (``'SE'``), z-score (``'z'``), and p-value (``'pvalue_raw'``) for each
        element in the input data frame.

    Raises
    ------
    To be added.
    """
    result = df.loc[:, ("score", "SE")]
    result["z"] = np.abs(result["score"] - score) / np.sqrt(result["SE"] ** 2 + se ** 2)
    result["pvalue_raw"] = 2 * stats.norm.sf(result["z"])
    return result



"""
Functions for identifying the type of
:py:class:`~enrich2.storemanager.StoreManager` derived object associated with a
given configuration object (decoded from a JSON file as described `here
<https://docs.python.org/2/library/json.html>`_).

"""


def is_experiment(cfg):
    """
    Check if the given configuration object specifies an
    :py:class:`~enrich2.experiment.Experiment`.

    Args:
        cfg (dict): decoded JSON object

    Returns:
        bool: True if `cfg` if specifies an
        :py:class:`~enrich2.experiment.Experiment`, else False.

    """
    if "conditions" in list(cfg.keys()):
        return True
    else:
        return False


def is_condition(cfg):
    """
    Check if the given configuration object specifies a
    :py:class:`~enrich2.condition.Condition`.

    Args:
        cfg (dict): decoded JSON object

    Returns:
        bool: True if `cfg` if specifies a
        :py:class:`~enrich2.condition.Condition`, else False.

    """
    if "selections" in list(cfg.keys()):
        return True
    else:
        return False


def is_selection(cfg):
    """
    Check if the given configuration object specifies a
    :py:class:`~enrich2.selection.Selection`.

    Args:
        cfg (dict): decoded JSON object

    Returns:
        bool: True if `cfg` if specifies a
        :py:class:`~enrich2.selection.Selection`, else False.

    """
    if "libraries" in list(cfg.keys()):
        return True
    else:
        return False


def is_seqlib(cfg):
    """
    Check if the given configuration object specifies a
    :py:class:`~enrich2.seqlib.SeqLib` derived object.

    Args:
        cfg (dict): decoded JSON object

    Returns:
        bool: True if `cfg` if specifies a :py:class:`~enrich2.seqlib.SeqLib`
        derived object, else False.

    """
    if "fastq" in list(cfg.keys()) or "identifiers" in list(cfg.keys()):
        return True
    else:
        return False


def seqlib_type(cfg):
    """
    Get the type of :py:class:`~enrich2.seqlib.SeqLib` derived object
    specified by the configuration object.

    Args:
        cfg (dict): decoded JSON object

    Returns:
        str: The class name of the :py:class:`~seqlib.seqlib.SeqLib` derived
        object specified by `cfg`.

    Raises:
        ValueError: If the class name cannot be determined.

    """
    if "barcodes" in cfg:
        if "map file" in cfg["barcodes"]:
            if "variants" in cfg and "identifiers" in cfg:
                raise ValueError("Unable to determine SeqLib type.")
            elif "variants" in cfg:
                return "BcvSeqLib"
            elif "identifiers" in cfg:
                return "BcidSeqLib"
            else:
                raise ValueError("Unable to determine SeqLib type.")
        else:
            return "BarcodeSeqLib"
    elif "overlap" in cfg and "variants" in cfg:
        return "OverlapSeqLib"
    elif "variants" in cfg:
        return "BasicSeqLib"
    elif "identifiers" in cfg:
        return "IdOnlySeqLib"
    else:
        raise ValueError("Unable to determine SeqLib type for configuration " "object.")


def element_type(cfg):
    """
    Get the type of :py:class:`~enrich2.storemanager.StoreManager` derived
    object specified by the configuration object.

    Args:
        cfg (dict): decoded JSON object

    Returns:
        str: The class name of the
        :py:class:`~enrich2.storemanager.StoreManager` derived object specified
        by `cfg`.

    Raises:
        ValueError: If the class name cannot be determined.

    """
    if is_experiment(cfg):
        return "Experiment"
    elif is_condition(cfg):
        return "Condition"
    elif is_selection(cfg):
        return "Selection"
    elif is_seqlib(cfg):
        return seqlib_type(cfg)
    else:
        raise ValueError("Unable to determine type for configuration object.")



from .storemanager import StoreManager
from .selection import Selection


class Condition(StoreManager):
    """
    Dummy class for experimental conditions within an 
    :py:class:`~experiment.Experiment`. Required for proper GUI behavior.
    """

    has_store = False  # don't create an HDF5 for Conditions
    treeview_class_name = "Condition"

    def __init__(self):
        StoreManager.__init__(self)
        self.selections = list()

    def configure(self, cfg, configure_children=True):
        StoreManager.configure(self, cfg)
        if configure_children:
            if "selections" not in cfg:
                raise KeyError(
                    "Missing required config value {} [{}]".format(
                        "selections", self.name
                    )
                )

            for sel_cfg in cfg["selections"]:
                sel = Selection()
                sel.configure(sel_cfg)
                self.add_child(sel)

    def serialize(self):
        """
        Format this object (and its children) as a config object suitable for dumping to a config file.
        """
        cfg = StoreManager.serialize(self)
        cfg["selections"] = [child.serialize() for child in self.children]
        return cfg

    def validate(self):
        """
        Calls validate on all child Selections.
        """
        for child in self.children:
            child.validate()

    def _children(self):
        """
        Method bound to the ``children`` property. Returns a list of all 
        :py:class:`~selection.Selection` objects belonging to this object, 
        sorted by name.
        """
        return sorted(self.selections, key=lambda x: x.name)

    def add_child(self, child):
        """
        Add a :py:class:`~selection.Selection`.
        """
        if child.name in self.child_names():
            raise ValueError(
                "Non-unique selection name '{}' [{}]".format(child.name, self.name)
            )
        child.parent = self
        self.selections.append(child)

    def remove_child_id(self, tree_id):
        """
        Remove the reference to a :py:class:`~selection.Selection` with 
        Treeview id *tree_id*.
        """
        self.selections = [x for x in self.selections if x.treeview_id != tree_id]
